"""Defines intermediary FormsiteParameters object"""

from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, Union
from datetime import datetime as dt
import pytz

# ----
from formsite_util.error import InvalidDateFormatExpection


def try_parse_date(date: Union[dt, str]) -> dt:
    """Attempts to parse a date string from various allowed formats int tz=utc datetime"""
    formats = ["%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%d", "%Y-%m-%d %H:%M:%S"]
    i = 0
    while not isinstance(date, dt):
        try:
            date = dt.strptime(date, formats[i])
            break
        except IndexError as ex:
            raise InvalidDateFormatExpection(date) from ex
        except ValueError:
            pass
        finally:
            i += 1
    return date


def shift_date_from_tz_to_utc(date: dt, tz: str) -> dt:
    """Converts a UTC date into UTC - tz_offset date"""
    offset = pytz.timezone(tz).utcoffset(date)
    return date + offset


@dataclass
class FormsiteParameters:

    """FormsiteParameters class
    This class stores parameters for Formsite requests.\n
    `after_id` gets only results with ID (Reference #) greater than integer you provide.\n
    `before_id` gets only results with ID (Reference #) less than integer you provide.\n
    `after_date` gets only results greater than input you provide, expects a `datetime object` or string in `ISO 8601`, `yyyy-mm-dd`, `yyyy-mm-dd HH:MM:SS`.\n
    `before_date` gets only results less than input you provide, expects a `datetime object` or string in `ISO 8601`, `yyyy-mm-dd`, `yyyy-mm-dd HH:MM:SS`.\n
    `timezone` sets the timezone dates in results are relative to, also affects input dates. Expects timezone name eg. `America/Chicago`. Defaults to `Etc/UTC`.\n
    `date_format` using python datetime directives specify what format you want your dates in your csv output file. Defaults to `%Y-%m-%d %H:%M:%S.`\n
    resultsview` More info on Formsite website or FS API of your specific form.\n
    `sort` ( "asc" | "desc" ) sorts results by reference number in ascending or descending order. Defaults to 'desc'.
    """

    last: Optional[int] = None
    after_id: Optional[int] = None
    before_id: Optional[int] = None
    after_date: Optional[Union[str, dt]] = None
    before_date: Optional[Union[str, dt]] = None
    timezone: Optional[str] = "Etc/UTC"
    resultsview: Optional[int] = 11
    sort: Optional[str] = "desc"

    def as_dict(self, single_page_limit: int = 500) -> dict:
        """Generates a parameters dictionary that is later passed to params= kw argument when making API calls.

        Args:
            single_page_limit (int, optional): Results per page limit, 500 is maximum amount. Defaults to 500.

        Returns:
            dict: params dict
        """
        results_params: dict[str, Union[str, int]] = dict()
        results_params["page"] = 1
        results_params["limit"] = single_page_limit
        if self.after_id is not None:
            results_params["after_id"] = self.after_id
        if self.before_id is not None:
            results_params["before_id"] = self.before_id
        if self.after_date is not None:
            date = try_parse_date(self.after_date)
            date = shift_date_from_tz_to_utc(date, self.timezone)
            results_params["after_date"] = date
        if self.before_date is not None:
            date = try_parse_date(self.after_date)
            date = shift_date_from_tz_to_utc(date, self.timezone)
            results_params["before_date"] = date
        if self.resultsview is not None:  # 11 = all items + statistics results view
            results_params["results_view"] = self.resultsview

        return results_params

    def copy(self) -> FormsiteParameters:
        """Returns a new instance of FormsiteParameters with the same values as this one"""
        return type(self)(
            self.last,
            self.after_id,
            self.before_id,
            self.after_date,
            self.before_date,
            self.timezone,
            self.resultsview,
            self.sort,
        )



"""Defines the synchronous Form data downloaded."""

import os
import re
import shutil
from typing import List, Tuple
from urllib.parse import urlparse
from requests import Session
from formsite_util.error import FormsiteFileDownloadException


FS_PREFIX_PAT = re.compile(r"((f|sig)-((\d+?)-)+)")


class DownloadStatus:
    """A simple object representing the download status"""

    def __init__(self, ok: bool, url: str, path: str, message: str = "") -> None:
        self.ok = ok
        self.message = message
        self.url = url
        self.path = path
        self.filename = os.path.basename(path)
        self.download_dir = self.path.rsplit("/", 1)[0]

    def __repr__(self):
        if self.message:
            return f"<FAILED {self.message}> {self.url}"
        else:
            return f"<OK> {self.url}"

    def isok(self):
        """Downloaded ok"""
        return self.ok

    def raise_for_status(self):
        """Raises `FormsiteFileDownloadException` if status not ok"""
        if not self.ok:
            raise FormsiteFileDownloadException(self.message)


def url_basename(url: str) -> str:
    """Extracts the filename from a url"""
    return os.path.basename(urlparse(url).path)


def truncate_filename(fn: str, max_len: int = 25, end_peek: int = 5) -> str:
    """Return a truncated version of the filename"""
    if fn(fn) > max_len:
        out = fn[: max_len - end_peek] + "…" + fn[-end_peek:]
    else:
        out = fn[:max_len]

    return out


def strip_prefix_filename(filename: str):
    """Removes `f-xxx-xxx-` prefix from a filename, leaving only ref_filename.ext"""
    return FS_PREFIX_PAT.sub(r"", filename)


def filter_urls(
    urls: List[str],
    download_dir: str,
    strip_prefix: bool = False,
    filename_substitution_re_pat: str = r"",
    overwrite_existing: bool = True,
) -> List[Tuple[str, str, str]]:
    """Filter list of URLs and filenames based on input filters. Returns list of (url, filename, path)"""
    filtered_URLs = []
    filename_dict = {}
    ls = os.listdir(download_dir)
    subsitution_pattern = re.compile(filename_substitution_re_pat)
    # 1st pass - filter URLs
    for url in urls:
        filename = url_basename(url)
        if strip_prefix:
            filename = strip_prefix_filename(filename)
        if filename_substitution_re_pat:
            filename = subsitution_pattern.sub("", filename)
        if not overwrite_existing and filename in ls:
            continue
        if filename not in filename_dict:
            filename_dict[filename] = []
        filename_dict[filename].append(url)
    # 2nd pass - find duplicate filenames
    for filename, urls in filename_dict.items():
        if len(urls) > 1:
            for i, url in enumerate(urls, start=1):
                filename_noext, ext = filename.rsplit(".", 1)
                new_filename = f"{filename_noext}_{i}.{ext}"
                filtered_URLs.append((url, f"{download_dir}/{new_filename}"))
        else:
            filtered_URLs.append((urls[0], f"{download_dir}/{filename}"))

    return filtered_URLs


def download_sync(
    url: str,
    path: str,
    session: Session,
    timeout: float = 160.0,
    max_attempts: int = 3,
) -> None:
    """Performs the download of 1 file to path. Performs error handling."""
    status = DownloadStatus(False, url, path)
    for attempt in range(1, max_attempts + 1):
        try:
            tmp_path = path + ".tmp"
            with session.get(url, stream=True, timeout=timeout) as resp:
                with open(tmp_path, "wb") as fp:
                    for content in resp.iter_content(chunk_size=4096):
                        fp.write(content)
            shutil.move(tmp_path, path)
            status.ok = True
            return status
        except Exception as ex:
            status.message = f"Failed {attempt} times | {ex}"
        finally:
            try:  # get rid of file.tmp
                os.remove(tmp_path)
            except Exception as _:
                pass
    return status



"""auth.py submodule, defined FormsiteCredentials class
Raises:
    ValueError: Invalid input when initializing FormsiteCredentials class.
"""

from dataclasses import dataclass
from typing import Dict


def _sanitize_argument(argument: str, chars2remove: Dict[str, str]) -> str:
    """Performs a find and replace on 'argument' based on mapping in chars2remove.

    Args:
        argument (str): text to perform find and replace on
        chars2remove (List[str]): mapping of {find:replace}

    Returns:
        str: sanitized argument
    """
    for key, value in chars2remove.items():
        argument = str(argument).replace(key, value)
    return argument


def _confirm_arg_format(arg_value: str, arg_name: str, flag: str, example: str) -> str:
    """A boiler plate function to display a helpful error message.

    Args:
        arg_value (str): current value of the argument variable
        arg_name (str): exact name of the argument variable
        flag (str): flag used to invoke the argument in terminal
        example (str): correct example value

    Raises:
        ValueError: Raised upon entering the incorrect argumant/format.

    Returns:
        str: the sanitized (quote-less) argument back
    """
    quotes_map = {"'": "", '"': ""}
    if not isinstance(arg_value, str):
        raise ValueError(
            f"invalid format for argument {arg_value}, {arg_name}, "
            f"correct example: {flag} {example}"
        )
    arg_value = _sanitize_argument(arg_value, quotes_map)
    return arg_value


@dataclass
class FormsiteCredentials:
    """Class representing formsite login information.

    Args:
        token (str): API access token

        server (str): formsite server, can be found in your formsite url at the beginning, https://fs_.(...).com

        directory (str): can also be found in your formsite URL, when accessing a specific form

    Returns:
        FormsiteCredentials instance
    """

    token: str
    server: str
    directory: str

    def __post_init__(self) -> None:
        """Confirms validity of input."""
        self.confirm_validity()

    def get_auth_header(self) -> dict:
        """Returns a dictionary sent as a header in the API request for authorization purposes."""
        return {"Authorization": f"bearer {self.token}", "Accept": "application/json"}

    def confirm_validity(self) -> None:
        """Checks if credentials input is in correct format."""
        self.token = _confirm_arg_format(self.token, "token", "-t", "token")
        self.server = _confirm_arg_format(self.server, "server", "-s", "fs1")
        self.directory = _confirm_arg_format(self.directory, "directory", "-d", "Wa37fh")



"""Defines utility functions for caching form items and form results."""

from typing import Optional
import json
import pandas as pd

from formsite_util.consts import METADATA_COLS


def items_load(path: str) -> Optional[dict]:
    """Attempts to load items from a file

    Args:
        path (str): Path where the items are stored

    Returns:
        Optional[dict]: Items dict or None if not found
    """
    try:
        with open(path, "r", encoding="utf-8") as fp:
            return json.load(fp)
    except FileNotFoundError:
        return None


def items_save(items: dict, path: str):
    """Saves items to a file

    Args:
        items (dict): Items dict
        path (str): Path where to store the items
    """
    with open(path, "w", encoding="utf-8") as fp:
        json.dump(items, fp, indent=4)


def items_match_data(items: dict, data_cols: pd.Index) -> bool:
    """Checks if all items column ids in data exist in items dict ids

    Args:
        items (dict): form.items
        data_cols (pd.Index): form.data.columns

    Returns:
        bool: True if they match each other, False if they do not.
    """
    metadata = set(METADATA_COLS.keys())
    results_ids = set(data_cols).difference(metadata)
    item_ids = set(i["id"] for i in items["items"])
    return len(results_ids.difference(item_ids)) == 0


def results_load(path: str) -> pd.DataFrame:
    """Loads data in a file located at `path` specified by serialization format (by file extension)

    Args:
        path (str): Path where the data is stored

    Supported formats are:
        - parquet
        - feather
        - pkl | pickle
        - hdf
    Raises:
        ValueError: In the event of an unsupported serialization format

    Returns:
        Optional[pd.DataFrame]: Loaded data as Pandas DataFrame. Empty if no data exists at specified path
    """
    ext = path.rsplit(".", 1)[-1].lower().strip()
    try:
        if ext == "parquet":
            df = pd.read_parquet(path)
        elif ext == "feather":
            df = pd.read_feather(path)
        elif ext in ("pkl" "pickle"):
            df = pd.read_pickle(path)
        elif ext == "xlsx":
            df = pd.read_excel(path)
        elif ext == "hdf":
            df = pd.read_hdf(path, key="data")
        else:
            raise ValueError(
                f"Invalid extension in path, '{ext}' is not a supported serialization format"
            )
    except FileNotFoundError:
        df = pd.DataFrame()
    except ValueError:  # For json
        df = pd.DataFrame()

    return df


def results_save(data: pd.DataFrame, path: str):
    """Saves data to a file in path in specified serialization format (by file extensions)

    Args:
        data (pd.DataFrame): Pandas DataFrame data
        path (str): Path where to store the data

    Supported formats are:
        - parquet
        - feather
        - pkl | pickle
        - hdf

    Raises:
        ValueError: In the event of unsupported serialization format
    """
    ext = path.rsplit(".", 1)[-1].lower().strip()

    if ext == "parquet":
        data.to_parquet(path)
    elif ext == "feather":
        data.to_feather(path)
    elif ext in ("pkl" "pickle"):
        data.to_pickle(path)
    elif ext == "xlsx":
        data.to_excel(path)
    elif ext == "hdf":
        data.to_hdf(path, key="data")
    else:
        raise ValueError(
            f"Invalid extension in path, '{ext}' is not a supported serialization format"
        )



"""Defines the FormFetcher object and its logic."""

from time import sleep
from typing import Generator
from math import ceil
from requests import Session, Response, HTTPError
from requests.exceptions import ConnectionError as requests_ConnectionError
from formsite_util.error import (
    FormsiteForbiddenException,
    FormsiteFormNotFoundException,
    FormsiteInternalException,
    FormsiteInvalidAuthenticationException,
    FormsiteInvalidParameterException,
    FormsiteRateLimitException,
)
from formsite_util.logger import FormsiteLogger
from formsite_util.parameters import FormsiteParameters
from formsite_util.consts import ConnectionError_DELAY, HTTP_429_WAIT_DELAY


class FormFetcher:
    """Performs API Interaction"""

    def __init__(
        self,
        form_id: str,
        token: str,
        server: str,
        directory: str,
        params: FormsiteParameters,
    ):
        """FormFetcher constructor

        Args:
            form_id (str): FormsiteForm ID
            token (str): Formsite API access token
            server (str): Formsite server (fsX.formsite. ...)
            directory (str): Forms directory
            params (FormsiteParameters): Results fetch parameters
        """
        self.form_id = form_id
        self.params = params
        self.results_params = params.as_dict()
        # ----
        self.url_base: str = f"https://{server}.formsite.com/api/v2/{directory}"
        self.url_results = f"{self.url_base}/forms/{self.form_id}/results"
        self.url_items = f"{self.url_base}/forms/{self.form_id}/items"
        self.auth_header = {"Authorization": f"bearer {token}"}
        # ----
        self.total_pages = 1
        self.cur_page = 1
        self.logger: FormsiteLogger = FormsiteLogger()

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.form_id}>"

    def fetch_iterator(self) -> Generator[dict, None, None]:
        """Iterates through all HTTP response pages

        Yields:
            Generator[dict]: data_of_page
        """
        max_page_sz = 500
        page_sz = self.results_params.get("limit", max_page_sz)
        with Session() as session:
            session.headers.update(self.auth_header)
            while True:
                try:
                    if self.cur_page > self.total_pages:
                        return StopIteration
                    resp = self.fetch_result(self.cur_page, session)
                    self.handle_response(resp)
                    self.total_pages = int(resp.headers.get("Pagination-Page-Last"))
                    self.total_pages = min(ceil(page_sz / max_page_sz), self.total_pages)
                    self.logger.debug(
                        f"Formsite API fetch {self.form_id} results | {self.cur_page}/{self.total_pages}"
                    )
                    self.cur_page += 1
                    yield resp.json()
                except FormsiteRateLimitException:
                    self.logger.debug(
                        f"Formsite API fetch reached RateLimitException | waiting {HTTP_429_WAIT_DELAY} seconds"
                    )
                    sleep(HTTP_429_WAIT_DELAY)

    def fetch_result(self, page: int, session: Session) -> Response:
        """Fetches a particular page of results

        Args:
            page (int): Current result page to fetch
            session (Session): Existing authorized session

        Returns:
            Response: Closed requests.Response object
        """
        params = self.results_params.copy()
        params["page"] = page
        try:
            with session.get(self.url_results, params=params) as resp:
                return resp
        except requests_ConnectionError:
            self.logger.critical(
                f"API Fetch {self.form_id}: Target refused connection, waiting and retrying"
            )
            sleep(ConnectionError_DELAY)
            return self.fetch_result(page, session)

    def fetch_items(self, results_labels_id: int = None) -> dict:
        """Fetches form items

        Args:
            results_labels_id (int, optional): Fetch result labels of this particular id. Defaults to None.

        Returns:
            dict: Formsite form's items dictionary
        """
        with Session() as session:
            session.headers.update(self.auth_header)
            params = (
                {"results_labels": results_labels_id}
                if results_labels_id is not None
                else {}
            )
            with session.get(self.url_items, params=params) as resp:
                self.handle_response(resp)
                self.logger.debug(f"Formsite API fetch: {self.form_id} items")
                return resp.json()

    @staticmethod
    def handle_response(response: Response):
        """Handles HTTP Repsonse code"""
        if response.status_code == 200:
            pass
        elif response.status_code == 401:  # Authentication info is missing or invalid.
            raise FormsiteInvalidAuthenticationException(
                response,
                "Please check if token, directory and server are correct",
            )
        elif response.status_code == 403:  # Forbidden.
            raise FormsiteForbiddenException(response.text)
        elif response.status_code == 404:  # Path or object not found.
            raise FormsiteFormNotFoundException(response.text)
        elif response.status_code == 422:  # Invalid parameter.
            raise FormsiteInvalidParameterException(response.text)
        elif response.status_code == 429:  # Too many requests or too busy.
            raise FormsiteRateLimitException(response.text)
        elif response.status_code >= 500:  # Unexpected Formsite internal error.
            raise FormsiteInternalException(response.text)
        else:
            raise HTTPError(response.text)



"""

## High level interfaces

FormsiteParameters: Represents parameters for results/items requests

FormsiteForm: Represents the form data and session

FormsiteFormsList: Represents the list of all forms for the specified account

FormCache: Store formsite forms in a folder

## Low level interfaces

FormFetcher: Result/Item fetching operations

FormParser: Result/Item parsing operations

FormData: Represents the form data without session

FormsiteLogger: Custom logger you may connect to your own logging

"""
__version__ = "2.0.2"

from .form import FormsiteForm
from .list import FormsiteFormsList
from .parameters import FormsiteParameters
from .form_fetcher import FormFetcher
from .form_parser import FormParser
from .form_data import FormData

# Legacy functionality
from .legacy import FormsiteCredentials, FormsiteInterface, FormsiteParams



"""
interfaces.py

`FormsiteInterface` `FormsiteParams` and `FormsiteCredentials` classes are defined here.
Author: Jakub Strnad
Documentation: https://github.com/strny0/formsite-utility
"""
from __future__ import annotations
import csv
import asyncio
from datetime import datetime as dt
from datetime import timedelta as td
from pathlib import Path
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Optional, Set, Union, Tuple, List
import re
import os
import pandas as pd
from pytz import UnknownTimeZoneError, timezone as pytztimezone
import requests
from .downloader import _FormsiteDownloader
from .processing import _FormsiteProcessing
from .api import _FormsiteAPI
from .auth import FormsiteCredentials
from tqdm import tqdm


def _shift_param_date(date: Union[str, dt], timezone_offset: td) -> str:
    """Shifts input date in the string format/datetime by timedelta in timezone offset.
    The offset is additive, date + timezone_offset.

    Args:
        date (Union[str, dt]): String in formats: 'yyyy-mm-dd', 'yyyy-mm-dd HH:MM:SS' or 'yyyy-mm-ddTHH:MM:SSZ' or datetime.
        timezone_offset (td): A timedelta value representing addition to 'date'.

    Notes:
        If you input time in ISO 8601 UTC, timezone offset won't be applied.

    Raises:
        ValueError: Raised if string format is not recognized.

    Returns:
        str: A datetime string in 'yyyy-mm-ddTHH:MM:SSZ' format, shifted by timezone_offset amount.
    """
    if isinstance(date, dt):
        date = date + timezone_offset
    else:
        formats = ["%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%d", "%Y-%m-%d %H:%M:%S"]
        for f in formats:
            try:
                date = dt.strptime(date, f)
                date = date + timezone_offset
                break
            except ValueError:
                continue
        if not isinstance(date, dt):
            raise ValueError(
                """invalid date format input for afterdate/beforedate, please use a datetime object or string in ISO 8601, yyyy-mm-dd or yyyy-mm-dd HH:MM:SS format"""
            )

    return dt.strftime(date, "%Y-%m-%dT%H:%M:%SZ")


def _extract_timezone_from_str(timezone: str, timezone_re: str) -> td:
    """Parses input timezone.

    Args:
        timezone (str): string in format ['+0200', '02:00', +02:00', '16:48', '-05:00', '-0600']
        timezone_re (str): regex encompassing above mentioned valid formats

    Returns:
        timedelta: timedelta offset
    """
    tz_offset = None
    if re.search(timezone_re, timezone) is not None:
        tz_str = timezone.replace(r"\'", "").replace(r"\"", "")
        if ":" in tz_str:
            tz_tuple = tz_str.split(":", 1)
        else:
            t = len(tz_str) - 2
            tz_tuple = (tz_str[:t], tz_str[-2:])

        tz_offset = td(hours=int(tz_tuple[0]), seconds=int(tz_tuple[1]) / 60)
    return tz_offset


def _calculate_tz_offset(timezone: str) -> Tuple[td, td, dt]:
    """Calculates timezone offset relative to input TARGET timezone string UTC.

    Args:
        timezone (str): String in format Ex. 'America/Chicago' (tz_database_name) or offset like '0400', '+0400', '-7:00', '-14:00'

    Raises:
        UnknownTimeZoneError: If input is not a valid tz_databse name or offset.

    Returns:
        Tuple[timedelta, timedelta, datetime]: tuple of offset_from_local time as a timedelta, offset from utc and local_time (datetime)
    """
    local_date = dt.now()
    utc_date = dt.utcnow()
    diff_local_utc = local_date - utc_date
    offset_utc = local_date - utc_date
    if timezone == "local":
        offset_local = td(seconds=0)
    else:
        tz_input_regex = r"(\+|\-|)([0-1]\d[0-5]\d|[0-1]\d\:[0-5]\d|\d\:[0-5]\d)"
        # parses the following formats:
        # America/Chicago
        # +1200
        # 1300
        # -1200
        # 8:00
        # -8:00
        offset_local = _extract_timezone_from_str(timezone, tz_input_regex)

        if offset_local is None:
            if re.search(r"\w+/\w+", timezone) is not None:
                off_local = pytztimezone(timezone).localize(local_date).strftime("%z")
                t = len(off_local) - 2
                l_inp = (off_local[:t], off_local[-2:])
                offset_utc = td(hours=int(l_inp[0]), seconds=int(l_inp[1]) / 60)
                offset_local = offset_utc - diff_local_utc
            else:
                raise UnknownTimeZoneError(timezone)

    return offset_local, offset_utc, local_date


def _validate_path(path: str) -> str:
    """Converts input path into POSIX format. Creates parent directories if necessary.

    Args:
        path (str): path to a file or a folder in any format

    Returns:
        str: path to a file or a folder in posix format
    """

    output_file = Path(path).resolve().absolute()
    if output_file.is_dir():
        os.makedirs(output_file.as_posix(), exist_ok=True)
    else:
        os.makedirs(output_file.parent.as_posix(), exist_ok=True)

    return output_file.as_posix()


@dataclass
class FormsiteParams:

    """LEGACY DO NOT USE

    Use FormsiteParameters with FormsiteForm instead.

    This class stores parameters for Formsite requests\n
    `afterref` gets only results greater than integer you provide\n
    `beforeref` gets only results less than integer you provide\n
    `afterdate` gets only results greater than input you provide, expects a `datetime object` or string in `ISO 8601`, `yyyy-mm-dd`, `yyyy-mm-dd HH:MM:SS`\n
    `beforedate` gets only results less than input you provide, expects a `datetime object` or string in `ISO 8601`, `yyyy-mm-dd`, `yyyy-mm-dd HH:MM:SS`\n
    `timezone` sets the timezone dates in results are relative to, also affects input dates. Expects either an offset string in format eg. `+06:00` or database name eg. `America/Chicago`\n
    `date_format` using python datetime directives specify what format you want your dates in your csv output file. Defaults to `%Y-%m-%d %H:%M:%S`\n
    `resultslabels` and `resultsview` More info on Formsite website or FS API of your specific form.\n
    `sort` ( "asc" | "desc" ) sorts results by reference number in ascending or descending order.
    """

    last: Optional[int] = None
    afterref: Optional[int] = None
    beforeref: Optional[int] = None
    afterdate: Optional[Union[str, dt]] = None
    beforedate: Optional[Union[str, dt]] = None
    timezone: Optional[str] = "local"
    resultslabels: Optional[int] = None
    resultsview: Optional[int] = 11
    sort: Optional[str] = "desc"

    def __post_init__(self):
        """Calls `_calculate_tz_offset` internal function."""
        self.timezone = "local" if self.timezone is None else self.timezone
        (
            self.timezone_offset,
            self.timezone_offset_utc,
            self.local_datetime,
        ) = _calculate_tz_offset(self.timezone)

    def get_params_as_dict(self, single_page_limit: int = 500) -> dict:
        """Generates a parameters dictionary that is later passed to params= kw argument when making API calls.

        Args:
            single_page_limit (int, optional): Results per page limit, 500 is maximum amount. Defaults to 500.

        Returns:
            dict: params dict
        """
        results_params: dict[str, Union[str, int]] = dict()
        results_params["page"] = 1
        results_params["limit"] = single_page_limit
        if self.afterref is not None:
            results_params["after_id"] = self.afterref
        if self.beforeref is not None:
            results_params["before_id"] = self.beforeref
        if self.afterdate is not None:
            parsed_afterdate = _shift_param_date(self.afterdate, self.timezone_offset)
            results_params["after_date"] = parsed_afterdate
        if self.beforedate is not None:
            parsed_beforedate = _shift_param_date(self.beforedate, self.timezone_offset)
            results_params["before_date"] = parsed_beforedate
        if self.resultsview is not None:  # 11 = all items + statistics results view
            results_params["results_view"] = self.resultsview

        return results_params

    def get_items_as_dict(self) -> dict:
        """Returns a dict that gets parsed as parameters by aiohttp when making a request."""
        if self.resultslabels is None:
            return {}
        elif isinstance(self.resultslabels, int):
            return {"results_labels": self.resultslabels}
        else:
            raise ValueError(f"resultslabels must be an int, got '{self.resultslabels}'")


@dataclass
class FormsiteInterface:
    """LEGACY DO NOT USE

    Use FormsiteForm instead

    Documentation: https://pypi.org/project/formsite-util/

    Author: https://github.com/strny0/formsite-utility

    Note: You cannot change form_id, you must reinstantiate the class.

    Args:
        form_id (str): ID of the target form.
        auth (FormsiteCredentials): Instance of the FormsiteCredentials class.
        params (FormsiteParams): Instance of the FormsiteParams class.
        display_progress (bool): Display tqdm progressbars. Defaults to True.

    Internal variables:
        `Data` (DataFrame | None): Stores results as dataframe.
        `Links` (Set[str] | None): Stores all formsite upload links in a set.
        `items` (Dict[str,str], None): RAW items json.
        `results` (List[Dict[str,str]] | list): RAW list of results jsons.
        `url_forms` (str): url for fetching info about all forms.
        `url_files` (str): url for downloading files.

    Methods of interest (and return types):
        `FetchResults` (None): stores results in self.Data of the instance of this class.
        `ReturnResults` (DataFrame): returns self.Results, if it's None, calls FetchResults first.
        `WriteResults` (None): writes the dataframe to a file, if it's None, calls FetchResults first.
        `ExtractLinks` (None): stores extracted links in self.Links of the instance of this class.
        `ReturnLinks` (List[str]): returns a tuple of all links.
        `WriteLinks` (None): writes them to a file.
        `ListAllForms` (DataFrame | None): lists all forms on formsite, output them to console or save them to a file.
        `ListColumns` (DataFrame | None): lists all columns and column IDs of a form you set the interface for.
        `DownloadFiles` (None): downloads all files submitted to the form to a folder you specify.
        `WriteLatestRef` (None): writes highest reference number in results to a file you specify.

    Returns:
        FormsiteInterface: An instance of the FormsiteInterface class.
    """

    form_id: str
    auth: FormsiteCredentials
    params: FormsiteParams = FormsiteParams()
    display_progress: bool = True

    def __post_init__(self):
        """Initializes internal variables.

        Internal variables:
            `Data` (DataFrame | None): Stores results as dataframe.
            `Links` (Set[str] | None): Stores all formsite upload links in a set.
            `items` (Dict[str,str], None): RAW items json.
            `results` (List[Dict[str,str]] | list): RAW list of results jsons.
            `url_forms` (str): url for fetching info about all forms.
            `url_files` (str): url for downloading files.
        """

        self.url_forms = (
            f"https://{self.auth.server}.formsite.com/api/v2/{self.auth.directory}/forms"
        )
        self.url_files = (
            f"https://{self.auth.server}.formsite.com/{self.auth.directory}/files/"
        )
        self.Data = None
        self.Links = None
        self.items = None
        self.results = []

    def __enter__(self):
        """Allows use of context managers."""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Allows use of context managers."""
        del self

    def fetch_raw(
        self,
        get_items: bool = True,
        get_results: bool = True,
        short_delay: float = 5.0,
        long_delay: float = 15.0,
    ) -> Tuple[dict, List[dict]]:
        """Performs calls to formsite REST API.

        Args:
            get_items (bool, optional): Whether to get items. Defaults to True.
            get_results (bool, optional): Whether to get results. Defaults to True.
            short_delay (float, optional): Delay between API calls. Defaults to 5.0.
            long_delay (float, optional): Delay between API calls when page > 3. Defaults to 15.0.

        Returns:
            Tuple[dict, List[dict]]: items_json, List of results_json
        """
        api_handler = _FormsiteAPI(
            form_id=self.form_id,
            params=self.params,
            auth=self.auth,
            display_progress=self.display_progress,
            short_delay=short_delay,
            long_delay=long_delay,
        )

        return api_handler.Start(get_items=get_items, get_results=get_results)

    def make_dataframe(
        self, items: dict, results: List[dict], use_resultslabels: bool = False
    ) -> pd.DataFrame:
        """Returns a pandas dataframe from received API data.

        Args:
            items (Dict[str, str]): raw items json from API.
            results (List[dict]): raw results json from API.
            use_resultslabels (bool, optional): True: Use question labels or resultslabels if available. False: Use column IDs and metadata names. Defaults to True.

        Returns:
            pd.DataFrame: DataFrame of the export.
        """
        sort_asc = self.params.sort != "desc"
        processing_handler = _FormsiteProcessing(
            items=items,
            results=results,
            timezone_offset=self.params.timezone_offset_utc,
            sort_asc=sort_asc,
            use_resultslabels=use_resultslabels,
            display_progress=self.display_progress,
            params_last=self.params.last,
        )
        return processing_handler.Process()

    def FetchResults(self, use_resultslabels: bool = True) -> None:
        """Fetches results from formsite API according to specified parameters.\n
        Updates the `self.Data` variable which stores the dataframe.

        Args:
            use_resultslabels (bool, optional): True: Use question labels or resultslabels if available. False: Use column IDs and metadata names. Defaults to True.

        Raises:
            HTTPError: When received HTTP response code other than 200 or 429.
        """
        assert bool(
            self.form_id
        ), f"You must pass form id when instantiating FormsiteCredentials('form_id', login, params=...) you passed '{self.form_id}'"
        if use_resultslabels:
            self.items, self.results = self.fetch_raw(get_items=True, get_results=True)
        else:
            self.items, self.results = self.fetch_raw(get_items=False, get_results=True)

        self.Data = self.make_dataframe(
            self.items, self.results, use_resultslabels=use_resultslabels
        )

    def ReturnResults(self, column_ids_as_labels: bool = False) -> pd.DataFrame:
        """Returns pandas dataframe of results.

        Args:
            column_ids_as_labels (bool, optional): Whether or not to use actual question labels or custom results labels for column names OR use just the column IDs/default metadata name.
            Defaults to False.

        Returns:
            pd.DataFrame: DataFrame of results in specified parameters.
        """
        if self.Data is None:
            self.FetchResults(use_resultslabels=column_ids_as_labels)
        return self.Data

    def _xtract(
        self,
        x: Any,
        pattern: re.Pattern,
        links_set: set = set(),
        links_filter_pattern: re.Pattern = None,
        pbar: tqdm = None,
    ):
        """Matches all formsite files links and appends them to links_set

        Args:
            x (Any): any element from self.Data
            pattern (re.Pattern): formsite files link pattern
            links_set (set, optional): output set of all links. Defaults to set().
            links_filter_pattern (re.Pattern, optional): regex to match links. Defaults to None.
        """
        _ = pbar.update(1) if pbar is not None else None
        try:
            for url_raw in pattern.findall(x):
                if "|" in url_raw:
                    for part in url_raw.split(" | "):
                        if links_filter_pattern.search(url_raw) is not None:
                            links_set.add(part)
                else:
                    if links_filter_pattern.search(url_raw) is not None:
                        links_set.add(url_raw)
        except Exception as ex:
            # print(ex)
            pass

    def ExtractLinks(self, links_filter_re: str = r".+") -> None:
        """Stores a set of links in `self.Links` of files saved on formsite servers, that were submitted to the specified form.

        Args:
            links_filter_re (str, optional): Only keep links that match the regex pattern string.. Defaults to r'.+'.
        """
        if self.Data is None:
            self.FetchResults()
        links_re = fr"(https\:\/\/{self.auth.server}\.formsite\.com\/{self.auth.directory}\/files\/.*)"
        url_pattern = re.compile(links_re)
        filter_pattern = re.compile(links_filter_re)
        self.Links = set()
        if self.display_progress:
            pbar = tqdm(desc="Extracting download links", leave=False, unit=" cells")
        else:
            pbar = None

        _ = self.Data.applymap(
            lambda x: self._xtract(
                x,
                pattern=url_pattern,
                links_set=self.Links,
                links_filter_pattern=filter_pattern,
                pbar=pbar,
            )
        )
        try:
            pbar.close()
        except AttributeError:
            pass

    def _human_friendly_filesize(self, number: int) -> str:
        """Converts a number (filesize in bytes) to more readable filesize with units."""
        reductions = 0
        while number >= 1024:
            number = number / 1024
            reductions += 1
        unit = {0: "", 1: "K", 2: "M", 3: "G", 4: "T", 5: "P", 6: "E"}
        return f"{number:0.2f} {unit.get(reductions, None)}B"

    def _list_all_forms(self) -> pd.DataFrame:
        """Internal function that performs the API fetch, data cleanup and formatting.

        Returns:
            pd.DataFrame: Cleaned DataFrame of all forms.
        """
        with requests.get(
            self.url_forms, headers=self.auth.get_auth_header()
        ) as response:
            response.raise_for_status()
            all_forms_json = response.json()["forms"]
            # un-nest the stats object
            for row in all_forms_json:
                for val in row["stats"]:
                    row[val] = row["stats"][val]
            for row in all_forms_json:
                for val in row["publish"]:
                    row[val] = row["publish"][val]
            forms_df = pd.DataFrame(
                all_forms_json,
                columns=[
                    "name",
                    "state",
                    "directory",
                    "resultsCount",
                    "filesSize",
                    "embed_code",
                    "link",
                ],
            )
            forms_df["form_id"] = forms_df.pop("directory")
            return forms_df

    def ListAllForms(
        self,
        sort_by: str = "name",
        display: bool = False,
        save2csv: Union[str, bool] = False,
    ) -> pd.DataFrame:
        """Prints name, id, results count, filesize and status of all forms into console or csv.
        You can sort in descending order by `name` `form_id` `resultsCount` `filesSize`.

        Args:
            sort_by (str, optional): One of {'name', 'form_id', 'resultsCount', 'filesSize'}. Defaults to 'name'.
            display (bool, optional): Print to stdout. Defaults to False.
            save2csv (Union[str, bool], optional): Path to output csv file. Defaults to False.

        Returns:
            pd.DataFrame: Dataframe with all relevant forms data.
        """
        forms_df = self._list_all_forms()
        if display:
            pd.set_option("display.max_rows", None)
            pd.set_option("display.max_columns", None)
            pd.set_option("display.width", None)
            pd.set_option("display.max_colwidth", 42)  # ensures width < 80 cols
            forms_df = forms_df[["name", "resultsCount", "filesSize", "form_id"]]
            forms_df.sort_values(
                by=[sort_by], inplace=True, ascending=False, ignore_index=True
            )
            forms_df.reset_index(drop=True, inplace=True)
            forms_df = forms_df.append(
                {
                    "name": "Total:",
                    "resultsCount": forms_df["resultsCount"].sum(),
                    "filesSize": forms_df["filesSize"].sum(),
                    "form_id": f"{forms_df.shape[0]} forms",
                },
                ignore_index=True,
            )
            forms_df["filesSize"] = forms_df["filesSize"].apply(
                lambda x: self._human_friendly_filesize(int(x))
            )
            forms_df.set_index("name", inplace=True)
            print(forms_df)

        if save2csv is not False:
            forms_df.sort_values(by=[sort_by], inplace=True, ascending=False)
            forms_df.set_index("name", inplace=True)
            output_file = _validate_path(str(save2csv))
            forms_df.to_csv(output_file, encoding="utf-8-sig")

        return forms_df

    def ReturnLinks(self, links_regex: str = r".+") -> Set[str]:
        """Stores a set of links in `self.Links` of files saved on formsite servers, that were submitted to the specified form.

        Args:
            links_filter_re (str, optional): Only keep links that match the regex pattern string.. Defaults to r'.+'.
        """
        if self.Links is None or links_regex != r".+":
            self.ExtractLinks(links_filter_re=links_regex)
        return self.Links

    def WriteLinks(
        self,
        destination_path: str,
        links_regex: str = r".+",
        sort_descending: bool = True,
    ) -> None:
        """Writes links extracted with `self.ExtractLinks()` to a .txt file.

        Args:
            destination_path (str): path to output file
            links_regex (str, optional): Include only links that match target regex. Defaults to r'.+'.
            sort_descending (bool, optional): Defaults to True.
        """
        if self.Links is None or links_regex != r".+":
            self.ExtractLinks(links_filter_re=links_regex)
        output_file = _validate_path(destination_path)
        sorted_links = [link + "\n" for link in self.Links]
        sorted_links.sort(reverse=sort_descending)
        with open(output_file, "w") as writer:
            writer.writelines(sorted_links)

    def DownloadFiles(
        self,
        download_folder: str,
        max_concurrent_downloads: int = 10,
        links_regex: str = r".+",
        filename_regex: str = r"",
        overwrite_existing: bool = True,
        report_downloads: bool = False,
        timeout: float = 80,
        retries: int = 1,
        strip_prefix: bool = False,
    ) -> None:
        """Downloads files saved on formsite servers, that were submitted to the specified form.
        Please customize `max_concurrent_downloads` to your specific use case.

        Args:
            `download_folder` (str): Path to target download folder
            `max_concurrent_downloads` (int, optional): Defaults to 10.
            `links_regex` (str, optional): Example: r'.+\\.jpg$' would get all files that end with .jpg. Defaults to r'.+'.
            `filename_regex` (str, optional): Removes characters that don't match regex from remote-filename. Defaults to r''.
            `overwrite_existing` (bool, optional): Whether or not to overwrite existing files. Defaults to True.
            `report_downloads` (bool, optional): Generates a report .txt file. Defaults to False.
            `timeout` (float, optional): In seconds, specify how long to wait for connection/download. Defaults to 80.
            `retries` (int, optional): In case of failed download or a timeout error, how many times to retry download. Defaults to 1.
            `strip_prefix` (bool, optional): If True, strips f-xxx-xxx prefix. Defaults to False.

        Raises:
            AssertationError: if len(self.Links) < 1, ie. there is nothing to download.
        """
        if self.Links is None:
            self.ExtractLinks(links_filter_re=links_regex)
        assert (
            len(self.Links) > 0
        ), f"There are no files to be downloaded for form {self.form_id}"

        download_folder = _validate_path(download_folder)
        download_handler = _FormsiteDownloader(
            download_folder,
            self.Links,
            max_concurrent_downloads,
            overwrite_existing=overwrite_existing,
            filename_regex=filename_regex,
            report_downloads=report_downloads,
            timeout=timeout,
            retries=retries,
            display_progress=self.display_progress,
            strip_prefix=strip_prefix,
        )
        asyncio.get_event_loop().run_until_complete(download_handler.Start())

    def ListColumns(self, print_stdout: bool = True) -> pd.DataFrame:
        """Prints list of columns (items, usercontrols) and their respective formsite IDs."""
        api_handler = _FormsiteAPI(
            form_id=self.form_id,
            params=self.params,
            auth=self.auth,
            display_progress=self.display_progress,
        )
        api_handler.check_pages = False
        items, _ = api_handler.Start(get_items=True, get_results=False)
        items = pd.DataFrame(items["items"], columns=["id", "label", "position"])
        if print_stdout:
            print("id : label")
            print("-----items----")
            items.apply(lambda row: print(f"{row['id']} : {row['label']}"), axis=1)
            print("---metadata---")
            print("id : Reference #")
            print("result_status : Status")
            print("date_start : Start Time")
            print("date_finish : Finish Time")
            print("date_update : Date")
            print("user_ip : User")
            print("user_browser : Browser")
            print("user_device : Device")
            print("user_referrer : Referrer")
            print("payment_status : Payment Status")
            print("payment_amount : Payment Amount Paid")
            print("login_username : Login Username")
            print("login_email : Login Email")
            print("--parameters--")
            print(f"Results labels: {self.params.resultslabels}")
            print(f"Results view: {self.params.resultsview}")
        return items

    def WriteResults(
        self,
        destination_path: str,
        encoding: str = "utf-8-sig",
        line_terminator: str = "os_default",
        separator: str = ",",
        date_format: str = "%Y-%m-%d %H:%M:%S",
        quoting: int = csv.QUOTE_MINIMAL,
    ) -> None:
        """Writes `self.Data` to a file based on provided extension.
        Supported output formats are (`.csv`|`.xlsx`|`.pkl`|`.pickle`|`.json`|`.parquet`|`.md`|`.txt`)

        Args:
            `destination_path` (str): path to output file
            `encoding` (str, optional): Defaults to "utf-8".
            `line_terminator` (str, optional): Defaults to 'os_default'.
            `separator` (str, optional): Defaults to ",".
            `date_format` (str, optional): Defaults to "%Y-%m-%d %H:%M:%S".
            `quoting` (int, optional): Pass values from csv.QUOTE_ enum from the python's csv module. Defaults to csv.QUOTE_MINIMAL.
        """
        if self.Data is None:
            self.FetchResults()
        output_file = _validate_path(destination_path)
        if re.search(".+\\.txt$", output_file) is not None:
            self.Data.to_string(output_file, encoding=encoding, index=False)
        elif (
            re.search(".+\\.pkl$", output_file) is not None
            or re.search(".+\\.pickle$", output_file) is not None
        ):
            self.Data.to_pickle(output_file)
        elif re.search(".+\\.parquet$", output_file) is not None:
            self.Data.to_parquet(output_file, index=False)
        elif re.search(".+\\.md$", output_file) is not None:
            self.Data.to_markdown(output_file, index=False)
        elif re.search(".+\\.json$", output_file) is not None:
            try:
                self.Data.to_json(output_file, orient="records", date_format="iso")
            except ValueError:
                renamer: dict[Any, Any] = defaultdict()
                for column_name in self.Data.columns[
                    self.Data.columns.duplicated(keep=False)
                ].tolist():
                    if column_name not in renamer:
                        renamer[column_name] = [column_name + "_0"]
                    else:
                        renamer[column_name].append(
                            column_name + "_" + str(len(renamer[column_name]))
                        )
                self.Data.rename(
                    columns=lambda column_name: renamer[column_name].pop(0)
                    if column_name in renamer
                    else column_name
                ).to_json(output_file, orient="records", date_format="iso")
        elif re.search(".+\\.xlsx$", output_file) is not None:
            print("Writing to excel (this can be slow for large files!)")
            self.Data.to_excel(output_file, index=False)
        else:
            line_terminator = (
                os.linesep if line_terminator == "os_default" else line_terminator
            )
            self.Data.to_csv(
                output_file,
                index=False,
                chunksize=1024,
                encoding=encoding,
                date_format=date_format,
                line_terminator=line_terminator,
                sep=separator,
                quoting=quoting,
            )

    def WriteLatestRef(self, destination_path: str) -> None:
        """Writes `max(self.Data['Reference #])` to a file.

        Args:
            `destination_path` (str): A Valid path to an output csv file.

        Raises:
            KeyError: If there is no 'id' or 'Reference #' column in specified results view.
        """
        if self.Data is None:
            self.FetchResults()

        output_file = _validate_path(destination_path)
        if "id" in self.Data.columns or "Reference #" in self.Data.columns:
            try:
                latest_ref = max(self.Data["Reference #"])
            except KeyError:
                latest_ref = max(self.Data["id"])
            with open(output_file, "w") as writer:
                writer.write(str(latest_ref))
        else:
            raise KeyError(
                "No valid column to extract Reference # from in specified parameters."
            )



"""Defines the FormData base class and its logic"""

from typing import Union
from pathlib import Path
import pandas as pd

# ----
from formsite_util.error import InvalidItemsStructureException
from formsite_util.logger import FormsiteLogger


class FormData:
    """Formsite API Form object, representing the data"""

    def __init__(self) -> None:
        """FormsiteFormData constructor"""
        self._labels: dict = None
        self._items: dict = None
        self._data: pd.DataFrame = pd.DataFrame()
        self.logger: FormsiteLogger = FormsiteLogger()

    @property
    def data(self) -> pd.DataFrame:
        """Formsite data as pandas DataFrame without items labels

        Returns:
            pd.DataFrame: form data
        """
        return self._data

    @data.setter
    def data(self, df) -> None:
        assert isinstance(df, pd.DataFrame), "Invalid value."
        self._data = df

    @data.deleter
    def data(self) -> None:
        del self._data

    @property
    def data_labels(self) -> pd.DataFrame:
        """Formsite data as pandas DataFrame with items labels

        Returns:
            pd.DataFrame: form data
        """
        if self.labels is None:
            return None
        else:
            return self._data.rename(columns=self.labels)

    @data_labels.setter
    def data_labels(self, *args, **kwargs):
        raise TypeError("Setting data for data_labels not allowed")

    @data_labels.deleter
    def data_labels(self):
        pass

    @property
    def labels(self) -> dict:
        """User defined labels

        Returns:
            Mapping of {id:label, ...}. None if unknown.
        """
        return self._labels

    @labels.setter
    def labels(self, value: dict):
        assert isinstance(value, dict), "Invalid value."
        self._labels = value

    @labels.deleter
    def labels(self):
        del self._labels

    @property
    def items(self) -> Union[list, None]:
        """Form's result labels object

        Items structure:
            Dictonary with object 'items' [...item...]
            Each item has an `id`, `label` and `position` records

        { "items": [
                {'id': '100', 'label': 'label_text', 'position': 1}, ...
            ]
        }

        Returns:
            List of records or None if not fetched.
        """

        return self._items

    @items.setter
    def items(self, value):
        if isinstance(value, dict) and "items" in value:
            self._items = value
        else:
            raise InvalidItemsStructureException(
                "Passed invalid items object to FormsiteForm,items or FormData.items. Expected a dictionary in the format {'items':[...]}"
            )

    @items.deleter
    def items(self):
        del self._items

    def to_csv(self, path: str, labels: bool = True, encoding: str = "utf-8-sig") -> None:
        """Save Formsite form as a csv with reasonable default settings"""
        path = Path(path).resolve().as_posix()
        df = self.data_labels if labels else self.data
        df.to_csv(
            path,
            date_format="%Y-%m-%d %H:%M:%S",
            index=False,
            encoding=encoding,
        )
        self.logger.debug(f"Form Data: Saved form to file '{path}'")

    def to_excel(self, path: str, labels: bool = True) -> None:
        """Save Formsite form as an excel with reasonable default settings (Warning: Slow for large data)"""
        path = Path(path).resolve().as_posix()
        df = self.data_labels if labels else self.data
        df.to_excel(path, index=False)
        self.logger.debug(f"Form Data: Saved form to file '{path}'")



"""Defines the CLI tool and its logic"""

import asyncio
import logging
import sys
import os
from pathlib import Path
from datetime import datetime as dt
from argparse import ArgumentParser, RawTextHelpFormatter, Namespace
from tqdm.auto import tqdm
import pandas as pd

# ----
from formsite_util.form import FormsiteForm
from formsite_util.list import FormsiteFormsList
from formsite_util.parameters import FormsiteParameters
from formsite_util.logger import FormsiteLogger
from formsite_util.consts import QUOTE, LINE_TERM, TIMESTAMP
from formsite_util.__init__ import __version__

_FETCH_PBAR = None
_DOWNLOAD_PBAR = None


def main():
    """The main program (CLI)"""
    global _FETCH_PBAR
    global _DOWNLOAD_PBAR
    log = FormsiteLogger()
    args = get_args()

    # Initialize logging
    if args.verbose:
        sh = logging.StreamHandler(sys.stdout)
        fmt = logging.Formatter("[%(levelname)s] %(message)s")
        sh.setFormatter(fmt)
        sh.setLevel(logging.DEBUG)
        log.addHandler(sh)
        args.disable_progressbars = True

    # Initialize session
    # ----
    if args.list_forms is not None:
        forms_list = FormsiteFormsList(args.token, args.server, args.directory)
        forms_list.fetch()
        # ----
        if not args.list_forms:
            if args.sort_list_by in ["results_count", "files_size"]:
                df = forms_list.data.sort_values(
                    by=args.sort_list_by,
                    ascending=False,
                )
            else:
                df = forms_list.data.sort_values(
                    by=args.sort_list_by,
                    ascending=True,
                )
            # ----
            df = df[["name", "form_id", "state", "results_count", "files_size_human"]]
            df.columns = ["name", "form_id", "state", "results_count", "files_size"]
            df.to_string(
                sys.stdout,
                index=False,
                max_colwidth=42,
            )
            print("")
        else:
            path = Path(args.list_forms).resolve().as_posix()
            forms_list.to_csv(path)
        sys.exit(0)
    # ----
    form = FormsiteForm(args.form, args.token, args.server, args.directory)
    params = FormsiteParameters(
        last=args.last,
        after_id=args.afterref,
        before_id=args.beforeref,
        after_date=args.afterdate,
        before_date=args.beforedate,
        resultsview=args.resultsview,
        timezone=args.timezone,
        sort=args.sort,
    )
    if not args.disable_progressbars:
        _FETCH_PBAR = tqdm(desc=f"Exporting {args.form}")
    form.fetch(
        params=params,
        fetch_callback=fetch_pbar_callback,
    )
    if not args.disable_progressbars:
        _FETCH_PBAR.close()
    # ----
    if args.output is not None:
        save_output(args, form)
    if args.latest_id is not None:
        save_latest_id(args, form)
    if args.extract is not None:
        save_extract(args, form)
    if args.download is not None:
        loop = asyncio.get_event_loop()
        if not args.disable_progressbars:
            _DOWNLOAD_PBAR = tqdm(desc=f"Downloading from {args.form}")
        save_download(args, form, loop)
        if not args.disable_progressbars:
            _DOWNLOAD_PBAR.close()
    # ----


def save_output(args: Namespace, form: FormsiteForm):
    """Save file based on extension."""
    # Supported (.csv|.xlsx|.pickle|.parquet|.feather|.hdf)
    if not args.output:
        path = Path(f"./export_{form.form_id}_{TIMESTAMP}.csv").resolve()
    else:
        path = Path(args.output).resolve()
    os.makedirs(path.parent.as_posix(), exist_ok=True)
    str_path = path.as_posix()
    ext = str_path.rsplit(".", 1)[-1].lower()
    df = form.data_labels if args.use_items else form.data
    if ext == "xlsx":
        # Write to excel with reasonable default settings
        df.to_excel(str_path, encoding="utf-8", index=False)
    elif ext in ("pickle", "pkl"):
        df.to_pickle(str_path)
    elif ext == "parquet":
        df.to_parquet(str_path)
    elif ext == "feather":
        df.to_feather(str_path)
    elif ext == "hdf":
        df.to_hdf(str_path, key=form.form_id)
    # Default to CSV
    else:
        df.to_csv(
            str_path,
            encoding=args.encoding,
            index=False,
            date_format=args.date_format,
            line_terminator=LINE_TERM.get(
                args.line_terminator, LINE_TERM.get("os_default")
            ),
            quoting=QUOTE[args.quoting],
            sep=args.separator,
        )


def save_latest_id(args: Namespace, form: FormsiteForm):
    """Write latest Reference # to a file (if it exists)"""
    m = None
    if form.uses_items is False and "id" in form.data.columns:
        m = max(form.data["id"])
    elif form.uses_items is True and "Reference #" in form.data.columns:
        m = max(form.data["Reference #"])

    if m is not None:
        with open(args.latest_id, "w", encoding="utf-8") as fp:
            fp.write(f"{m}\n")


def save_extract(args: Namespace, form: FormsiteForm):
    """Extract all URLs from the FileUpload controls and save them to a file"""

    if not args.extract:
        path = Path(f"./url_{form.form_id}_{TIMESTAMP}.txt").resolve()
    else:
        path = Path(args.extract).resolve()
    os.makedirs(path.parent.as_posix(), exist_ok=True)
    str_path = path.as_posix()
    URLs = form.extract_urls(args.extract_regex)
    with open(str_path, "w", encoding="utf-8") as fp:
        for url in URLs:
            fp.write(f"{url}\n")


def save_download(args: Namespace, form: FormsiteForm, loop: asyncio.AbstractEventLoop):
    """Download all files uploaded to the form using the File Upload control"""

    if not args.download:
        path = Path(f"./download_{form.form_id}_{TIMESTAMP}").resolve()
    else:
        path = Path(args.download).resolve()
    # ----
    os.makedirs(path.parent.as_posix(), exist_ok=True)
    str_path = path.as_posix()
    download = form.async_downloader(
        str_path,
        max_concurrent=args.concurrent_downloads,
        timeout=args.timeout,
        max_attempts=args.retries,
        url_filter_re=args.extract_regex,
        filename_substitution_re_pat=args.download_regex,
        overwrite_existing=args.dont_overwrite_downloads,
        strip_prefix=args.strip_prefix,
        callback=download_pbar_callback,
    )
    # ----
    loop.run_until_complete(download.run())


def fetch_pbar_callback(page: int, total_pages: int, data: dict) -> None:
    """Updates fetch progress bar (if enabled)"""
    if _FETCH_PBAR is None:
        return
    _FETCH_PBAR.total = total_pages
    _FETCH_PBAR.update(1)


def download_pbar_callback(url: str, filepath: str, total_files: int) -> None:
    """Updates download progress bar (if enabled)"""
    if _DOWNLOAD_PBAR is None:
        return
    _DOWNLOAD_PBAR.total = total_files
    _DOWNLOAD_PBAR.update(1)


def get_args() -> Namespace:
    """Uses argparse module to retrive argv arguments"""
    parser = ArgumentParser(
        prog="formsite-util",
        description="Github of author:\n"
        "https://github.com/strny0/formsite-utility\n"
        "This program performs an export of a specified formsite form with provided\nparameters.\n"
        "A faster alternative to a manual export from the formsite website, that can\nbe used for workflow automation.\n"
        "Allows download of files uploaded to the form.",
        epilog="More info can be found at Formsite API v2 help page:\n"
        "https://support.formsite.com/hc/en-us/articles/360000288594-API\n"
        "You can find API related information of a specific form under:"
        "[Form Settings > Integrations > Formsite API]\n"
        f"formsite-util  Copyright (C) {dt.now().year} Jakub Strnad\n"
        "This program comes with ABSOLUTELY NO WARRANTY; for details see LICENSE.md\n"
        "This is free software, and you are welcome to redistribute it.",
        formatter_class=RawTextHelpFormatter,
    )

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}\n",
    )
    g_auth = parser.add_argument_group("Authorization")
    g_params_r = parser.add_argument_group("Results Parameters")
    g_params_i = parser.add_argument_group("Items Parameters")
    g_output = parser.add_argument_group("Output file")
    g_extract = parser.add_argument_group("Extract links")
    g_download = parser.add_argument_group("Download files")
    g_other = parser.add_argument_group("Other functions")
    g_debug = parser.add_argument_group("Debugging")
    g_metadata = parser.add_mutually_exclusive_group(required=False)

    g_auth.add_argument(
        "-t",
        "--token",
        type=str,
        default=None,
        required=True,
        help="Your Formsite API token.",
    )
    g_auth.add_argument(
        "-s",
        "--server",
        type=str,
        default=None,
        metavar="FORMSITE_SERVER",
        required=True,
        help="Your Formsite server. A part of the url. https://fsX.forms… <- the 'fsX' part."
        "\nFor example 'fs22'.",
    )
    g_auth.add_argument(
        "-d",
        "--directory",
        type=str,
        default=None,
        metavar="FORMSITE_DIRECTORY",
        required=True,
        help="Your Formsite directory. Found under [Share > Links > Directory].",
    )
    g_params_r.add_argument(
        "-f",
        "--form",
        type=str,
        default=None,
        metavar="FORM_ID",
        help="Your Formsite form ID. Found under [Share > Links > Directory].",
    )
    g_params_r.add_argument(
        "--last",
        type=int,
        default=None,
        metavar="NUMBER",
        help="Gets last x results.",
    )
    g_params_r.add_argument(
        "-aID",
        "--afterref",
        type=int,
        default=None,
        metavar="NUMBER",
        help="Get results greater than a specified Reference #.",
    )
    g_params_r.add_argument(
        "-bID",
        "--beforeref",
        type=int,
        default=None,
        metavar="NUMBER",
        help="Get results lesser than a specified Reference #.",
    )
    g_params_r.add_argument(
        "-aDate",
        "--afterdate",
        type=str,
        default=None,
        metavar="DATETIME_STRING",
        help="Get results after a specified date."
        "\nMust be formatted as ISO 8601 UTC, YYYY-MM-DD, or YYYY-MM-DD HH:MM:SS."
        "\nThis date is in your local timezone, unless specified otherwise.",
    )
    g_params_r.add_argument(
        "-bDate",
        "--beforedate",
        type=str,
        default=None,
        metavar="DATETIME_STRING",
        help="Get results before a specified date."
        "\nMust be formatted as ISO 8601 UTC, YYYY-MM-DD, or YYYY-MM-DD HH:MM:SS."
        "\nThis date is in your local timezone, unless specified otherwise.",
    )
    g_params_r.add_argument(
        "--sort",
        choices=["asc", "desc"],
        type=str,
        default="desc",
        help="Determines how the output CSV will be sorted. Defaults to descending.",
    )
    g_params_i.add_argument(
        "--resultslabels",
        type=int,
        default=None,
        metavar="NUMBER",
        help="Use specific results labels for your CSV headers.\n"
        "Defaults to default Question labels.",
    )
    g_params_r.add_argument(
        "--resultsview",
        type=int,
        default=None,
        metavar="NUMBER",
        help="Extract a subset of columns with a resultsview ID. Defaults to 11 (All items + Metadata).",
    )
    g_params_r.add_argument(
        "-T",
        "--timezone",
        default="Etc/UTC",
        help="Specify the timezone relative to which you want your results.\n"
        "By default, results come in UTC.\n"
        "Input name of the timezone from this list:\n"
        "https://en.wikipedia.org/wiki/List_of_tz_database_time_zones\n"
        "[Examples:]\n"
        "'America/Chicago'\n"
        "'Europe/Paris'\n"
        "'America/New_York'\n"
        "'Asia/Bangkok'",
    )
    g_params_r.add_argument(
        "--delay",
        type=int,
        metavar="NUMBER",
        default=3,
        help="Delay in seconds between each API call. Defaults to 3 seconds.",
    )
    g_output.add_argument(
        "-o",
        "--output",
        nargs="?",
        default=None,
        metavar="PATH/TO/FILE",
        const="",
        help="Specify output file name and location.\n"
        "Defaults to export_yyyymmdd_formID.csv in the folder of the script.\n"
        "Supported output formats are (.csv|.xlsx|.pickle|.parquet|.feather|.hdf)",
    )
    g_output.add_argument(
        "-i",
        "--use_items",
        action="store_false",
        default=True,
        help="Instead of regular headers uses default metadata and items ids",
    )
    g_output.add_argument(
        "--encoding",
        type=str,
        default="utf-8-sig",
        help="Specify encoding of the output file (if output format supports it).\n"
        "More info on possible values here:\n"
        "https://docs.python.org/3/library/codecs.html#standard-encodings\n"
        "Defaults to 'utf-8-sig', UTF-8 with BOM.",
    )
    g_output.add_argument(
        "--separator",
        type=str,
        default=",",
        help="Specify separator of the output file (if output format supports it).\n"
        "Defaults to ',' (comma)",
    )
    g_output.add_argument(
        "--line_terminator",
        type=str,
        choices={"LF", "CR", "CRLF", "os_default"},
        default="os_default",
        help="Specify line terminator of the output file (if output format supports it)\n"
        "Defaults to 'LF'.",
    )
    g_output.add_argument(
        "--quoting",
        type=str,
        choices={"QUOTE_MINIMAL", "QUOTE_ALL", "QUOTE_NONNUMERIC", "QUOTE_NONE"},
        default="QUOTE_MINIMAL",
        help="Specify quoting level of the output file (if output format supports it).\n"
        "Defaults to 'QUOTE_MINIMAL'\n"
        "More info about the quoting levels: https://docs.python.org/3/library/csv.html\n",
    )
    g_output.add_argument(
        "--date_format",
        default="%Y-%m-%d %H:%M:%S",
        metavar="'POSIX DATETIME DIRECTIVE'",
        help="Specify datetime format in output file using python strftime directives.\n"
        "Defaults to '%%Y-%%m-%%d %%H:%%M:%%S' which is yyyy-mm-dd HH:MM:SS\n"
        "You can find the possible format directives here:\n"
        "https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior",
    )
    g_extract.add_argument(
        "-x",
        "--extract",
        nargs="?",
        default=None,
        metavar="PATH/TO/FILE",
        const="",
        help="Outputs a .txt file with all formsite files urls in specified form."
        "\nYou can specify file name or location, for example:"
        "\n'-x C:\\Users\\MyUsername\\Desktop\\download_links.txt'."
        "\nIf you don't specify a location, it will default to the folder of the script.",
    )
    g_download.add_argument(
        "-D",
        "--download",
        nargs="?",
        default=None,
        metavar="PATH/TO/FILE",
        const="",
        help="If you include this flag, all formsite links in the export will be downloaded to a folder."
        "\nYou can specify location, for example `-D 'C:\\Users\\My Username\\Desktop\\downloads'"
        "\nIf you don't specify a location, it will default to the folder of the script.",
    )
    g_other.add_argument(
        "-S",
        "--latest_id",
        nargs="?",
        metavar="PATH/TO/FILE",
        default=None,
        const="",
        help="If you enable this option, a text file `latest_ref.txt` will be created."
        "\nThis file will only contain the highest reference number in the export."
        "\nIf there are no results in your export, nothing will happen."
        "\nYou may also specify an output file, `-S output_file.txt",
    )
    g_extract.add_argument(
        "-xre",
        "--extract_regex",
        type=str,
        metavar="'REGEX'",
        default=r".+",
        help="Keep only links that match the regex you provide."
        "\nWon't do anything if -x or -d arguments are not provided."
        "\nDefaults to '.+'. Example usage: '-X \\.json$' would only give you files that have .json extension.",
    )
    g_download.add_argument(
        "-c",
        "--concurrent_downloads",
        default=5,
        metavar="NUMBER",
        type=int,
        help="You can specify the number of concurrent download tasks."
        "\nMore for large numbers of small files, less for large files."
        "\nDefault is 5.",
    )
    g_download.add_argument(
        "-n",
        "--dont_overwrite_downloads",
        default=True,
        action="store_false",
        help="Checks if files already exist in download directory based on filename."
        "If they already exist, they are not re-downloaded.",
    )
    g_download.add_argument(
        "-Dre",
        "--download_regex",
        default=r"",
        metavar=R"'REGEX'",
        type=str,
        help="If you include this argument, filenames of the files you download from formsite "
        "\nservers will remove all characters from their name that dont match the regex."
        "\nExpecting an input of allowed characters, for example: -R '[^\\w\\_\\-]+'"
        "\nAny files that would be overwritten as a result of the removal of characters will be appended with _1, _2, etc.",
    )
    g_download.add_argument(
        "--timeout",
        nargs=1,
        default=80,
        type=int,
        help="Timeout in seconds for each individual file download."
        "\nIf download time exceeds it, it will throw a timeout error and retry up until retries. Defaults to 80.",
    )
    g_download.add_argument(
        "--retries",
        nargs=1,
        default=3,
        type=int,
        help="Number of times to retry downloading files if the download fails. Defaults to 1.",
    )
    g_download.add_argument(
        "--strip_prefix",
        default=False,
        action="store_true",
        help="If you enable this option, filenames of downloaded files will not have f-xxx-xxx- prefix.",
    )

    g_metadata.add_argument(
        "-l",
        "--list_forms",
        nargs="?",
        default=None,
        const="",
        help="By itself, prints all forms, their form ids and status. You can specify a file to save the data into.\n"
        "Example: '-L ./list_of_forms.csv' to output to file or '-L' by itself to print to console.\n"
        "Requires login info. Overrides all other functionality of the program.",
    )
    g_other.add_argument(
        "-Q",
        "--sort_list_by",
        default="name",
        choices={"name", "results_count", "files_size"},
        help="You may chose what to sort -l commands by in descending order. Defaults to name.",
    )
    g_other.add_argument(
        "-P",
        "--disable_progressbars",
        action="store_true",
        default=False,
        help="If you use this flag, program will not display progressbars to console.",
    )

    g_debug.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Enable verbose logging to a log file.",
    )

    args, uargs = parser.parse_known_args()

    if uargs:  # if any unknown arguments were passed
        print(f"Unknown arguments: {uargs}")
        sys.exit(1)

    return args


if __name__ == "__main__":
    main()



"""Defines FormsiteFormsList object and its logic."""

from __future__ import annotations
from pathlib import Path
import pandas as pd
from requests import Session
from formsite_util.form_fetcher import FormFetcher
from formsite_util.logger import FormsiteLogger


def readable_filesize(number: int) -> str:
    """Converts a number (filesize in bytes) to more readable filesize with units."""
    if number is None:
        return None
    reductions = 0
    while number >= 1024:
        number = number / 1024
        reductions += 1
    unit = {0: "", 1: "K", 2: "M", 3: "G", 4: "T", 5: "P", 6: "E"}
    return f"{number:0.2f} {unit.get(reductions, None)}B"


class FormsiteFormsList:
    """Formsite API object, representing list of all forms in a directory"""

    def __init__(
        self,
        token: str,
        server: str,
        directory: str,
    ) -> None:
        """FormsiteFormsList constructor

        Args:
            token (str): Formsite API Token
            server (str): Formsite Server (fsX.formsite.com)
            directory (str): Formsite Directory
        """
        super().__init__()
        self.auth_header = {"Authorization": f"bearer {token}"}
        self._data: pd.DataFrame = pd.DataFrame()
        self.logger: FormsiteLogger = FormsiteLogger()
        self.url_base: str = f"https://{server}.formsite.com/api/v2/{directory}"
        self.url_forms: str = f"{self.url_base}/forms"

    @property
    def data(self) -> pd.DataFrame:
        """Formsite data as pandas DataFrame

        Returns:
            pd.DataFrame: form data
        """
        return self._data

    @data.setter
    def data(self, value):
        assert isinstance(value, pd.DataFrame), "Invalid value."
        self._data = value

    @data.deleter
    def data(self):
        del self._data

    def fetch(self):
        """Perform the API Fetch for the list of forms"""
        # GET https://{server}.formsite.com/api/v2/{user_dir}/forms
        with Session() as session:
            session.headers.update(self.auth_header)

            with session.get(self.url_forms) as resp:
                FormFetcher.handle_response(resp)
                data = resp.json()

        self.data = self.parse(data)

    def parse(self, data: dict) -> pd.DataFrame:
        """Parses forms list json into pandas dataframe"""

        rows = []
        for item in data["forms"]:
            row = {
                "form_id": item.get("directory", ""),
                "name": item.get("name", ""),
                "state": item.get("state", ""),
                "results_count": item["stats"].get("resultsCount", 0),
                "files_size": item["stats"].get("filesSize", 0),
                "files_size_human": readable_filesize(item["stats"].get("filesSize", 0)),
                "url": item["publish"].get("link", ""),
            }
            rows.append(row)

        return pd.DataFrame(rows)

    def to_csv(self, path: str, encoding: str = "utf-8-sig") -> None:
        """Save Formsite forms list as a csv with reasonable default settings"""
        path = Path(path).resolve().as_posix()
        self.data.to_csv(
            path,
            index=False,
            encoding=encoding,
        )

    def to_excel(self, path: str) -> None:
        """Save Formsite forms list as an excel with reasonable default settings"""
        path = Path(path).resolve().as_posix()
        self.data.to_excel(path, index=False)



"""Definitions of all custom exception classes."""

from requests import HTTPError


class FormsiteNoResultsException(Exception):
    """No results in specified parameters"""


class InvalidDateFormatExpection(Exception):
    """Entered invalid date format for after_date or before_date parameter"""


class FormsiteInvalidAuthenticationException(HTTPError):
    """Invalid token, server or directory"""


class FormsiteFormNotFoundException(HTTPError):
    """Invalid form id"""


class FormsiteInvalidParameterException(HTTPError):
    """Invalid parameter passed to request"""


class FormsiteRateLimitException(HTTPError):
    """Reached Formsite API rate limit, please wait at least 50 seconds"""


class FormsiteInternalException(HTTPError):
    """Unexpected error on FormSite servers (HTTP 5xx)"""


class FormsiteForbiddenException(HTTPError):
    """HTTP 403 Error"""


class FormsiteFileDownloadException(HTTPError):
    """Exception occured while downloading a file from formsite form"""


class FormsiteUncachableParametersException(Exception):
    """FormsiteForm was fetched with use_items=True or not fetched at all"""


class InvalidItemsStructureException(Exception):
    """FormsiteForm or FormData .items property is set incorrectly"""



"""Defines the FormParser object and its logic."""

from itertools import chain
import re
import pandas as pd

from formsite_util.logger import FormsiteLogger
from formsite_util.consts import METADATA_COLS


def _parse_item(item: dict):
    """Parses each item in results['results']['items'] to the export format"""
    if "value" in item:
        value = item["value"]
    elif "values" in item:
        values = sorted(item["values"], key=lambda x: x["position"])
        value = " | ".join(v["value"] for v in values)
    return value


def _parse_date_col_inplace(df: pd.DataFrame, col: str):
    """Tries to parse date column (string to datetime) inplace"""
    if col in df.columns:
        df[col] = pd.to_datetime(df[col], errors="raise")


def _order_df_cols(df: pd.DataFrame):
    """Reorders existing dataframe columns into standard export order"""
    final_order = []
    hardcoded_cols = list(METADATA_COLS.keys())
    df_cols = df.columns
    items_cols = set(df_cols).difference(set(hardcoded_cols))
    # ---- left side ----
    left_side = ["id", "result_status", "login_username", "login_email"]
    final_order += [col for col in left_side if col in df_cols]
    # ---- items ----
    final_order += [col for col in df_cols if col in items_cols]
    # ---- right side ----
    final_order += [
        col for col in hardcoded_cols if col in df_cols and col not in final_order
    ]
    # ----
    return df[final_order]


class FormParser:
    """Parses result json into a pandas dataframe"""

    def __init__(self) -> None:
        self.data = []
        self.children_item_re = re.compile(r"(\d+?-\d+?-\d+?)")
        self.logger: FormsiteLogger = FormsiteLogger()

    def parse_results_items(self, items: dict) -> dict:
        """Parses ['items'] dictionary of the result"""
        parsed = {}
        for item in items:
            key = item["id"]
            val = _parse_item(item)
            if self.children_item_re.match(key) is not None:
                s = key.split("-")
                parent_key = f"{s[0]}-{s[-1]}"
                if parent_key not in parsed:
                    parsed[parent_key] = val
                else:
                    parsed[parent_key] = f"{parsed[parent_key]} | {val}"
            else:
                parsed[key] = val
        self.logger.debug(f"Form Parser: Parsed {len(parsed)} item columns")
        return parsed

    def feed(self, results: dict) -> None:
        """Parses 1 Formsite results dictionary, appends it to processed data"""
        mdc = set(METADATA_COLS.keys())
        for record in results["results"]:
            keyset = set(record.keys())
            metadata = {i: record.get(i) for i in keyset.intersection(mdc)}
            items = self.parse_results_items(record.get("items", {}))
            self.data.append(dict(**metadata, **items))

    def as_dataframe(self) -> pd.DataFrame:
        """Return data fed into the parser so far as a Pandas DataFrame"""
        df = pd.DataFrame(self.data)
        if not df.empty:
            df = _order_df_cols(df)
            _parse_date_col_inplace(df, "date_update")
            _parse_date_col_inplace(df, "date_start")
            _parse_date_col_inplace(df, "date_finish")
        return df

    def as_records(self) -> pd.DataFrame:
        """Return data fed into the parser so far as a 1 json records object"""
        return list(chain(self.data))

    @staticmethod
    def create_rename_map(items: dict) -> dict:
        """Creates a mapping for pd.rename(column=...) from items (list of records {id, label, position}."""
        items_l = items["items"]
        rename_map = {}
        parent_label_map = {}
        label_map = {item["id"]: item["label"] for item in items_l}
        for item in items_l:
            key = item["id"]
            if "children" in item:
                for c in item["children"]:
                    parent_label_map[c] = item["label"]
            if key in parent_label_map:
                parent_label = parent_label_map.get(key)
                children = item.get("children")
                if children:
                    idx = int(key.split("-")[-1])
                    if idx < len(children):
                        c = children[int(idx)]
                        prev_label = label_map.get(c)
                        rename_map[key] = f"{prev_label} ({parent_label})"
                else:
                    rename_map[key] = f"{item['label']} ({parent_label})"
            else:
                rename_map[key] = item["label"]
        rename_map.update(METADATA_COLS)
        return rename_map



"""Defines a singleton logger object used by formsite-util."""


import logging


class FormsiteLogger(logging.Logger):
    """Custom logger. Key=formsite"""

    _instance = None
    _init_flag = False

    def __new__(cls, *args, **kwargs):
        """Prevents duplicate instances of this object"""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        if FormsiteLogger._init_flag:
            return

        super().__init__("formsite", logging.DEBUG)
        FormsiteLogger._init_flag = True



"""
api.py

This module contains a class that handles API requests to Formsite API.
"""
from __future__ import annotations
from typing import List, Tuple, Any
from dataclasses import dataclass
import time
import requests
from requests.exceptions import HTTPError
from tqdm import tqdm

from .auth import FormsiteCredentials


@dataclass
class _FormsiteAPI:
    """Handles retreival of results from your formsite form. Invoked with `self.Start()` method.

    Keywords:
        Results: Data from your formsite form (the rows).
        Items:   Labels of user-defined columns in your formsite form (the header row).

    Args:
        form_id (str): ID of form to fetch.
        params (FormsiteParams): An instance of FormsiteParams class.
        auth (FormsiteCredentials): An instance of FormsiteCredentials class.
        display_progress (bool): Display progress using tqdm. Defaults to True.
        delay (float | int): Delay in seconds between each API call. Defaults to 5.
        long_delay (float | int): Delay in seconds between each API call after page 3. Defaults to 15.

    Raises:
        Exception: General uncaught exception.
        HTTPError: Raised for all HTTP response codes other than 200 or 429.

    Returns:
        _FormsiteAPI: Instance of the _FormsiteAPI class. Start requests in `.Start()` method.
    """

    form_id: str
    params: Any
    auth: FormsiteCredentials
    display_progress: bool = True
    short_delay: float = 5.0
    long_delay: float = 15.0

    def __post_init__(self):
        """Generates post_init internal variables."""
        self.url_base: str = (
            f"https://{self.auth.server}.formsite.com/api/v2/{self.auth.directory}"
        )
        self.results_url: str = f"{self.url_base}/forms/{self.form_id}/results"
        self.items_url: str = f"{self.url_base}/forms/{self.form_id}/items"
        self.total_pages: int = 1
        self.check_pages: bool = True
        self.pbar = (
            tqdm(
                desc="Starting API requests",
                total=2,
                unit="call",
                leave=False,
                dynamic_ncols=True,
                ncols=80,
            )
            if self.display_progress
            else None
        )
        self.session = None
        self.params_dict = self.params.get_params_as_dict()
        self.items_dict = self.params.get_items_as_dict()

    def Start(self, get_items: bool, get_results: bool) -> Tuple[dict, List[dict]]:
        """Performs all API calls to formsite servers asynchronously"""
        items, results = (None, [])
        with requests.session() as self.session:
            self.session.headers.update(self.auth.get_auth_header())
            results = self.get_results() if get_results else []
            items = self.fetch_items() if get_items else None
            self._update_pbar_progress()
            self._update_pbar_desc(desc="API calls complete")
            try:
                self.pbar.close()
            except AttributeError:
                pass
        return items, results

    def get_results(self) -> List[dict]:
        """Fetches all results that match input parameters from a form.

        Raises:
            HTTPError

        Returns:
            List[dict]: List of pages of results.
        """
        results = [self.fetch_results(self.params_dict, 1)]
        if self.total_pages > 1:
            for page in range(2, self.total_pages + 1):
                try:
                    if isinstance(self.params.last, int):
                        if (page * self.params_dict["limit"]) >= self.params.last:
                            self._update_pbar_total(page)
                            break
                    results.append(self.fetch_results(self.params_dict, page))
                except Exception as err:
                    print("\r\n")
                    print(err)
                    try:
                        del self.params_dict["after_id"]
                    except KeyError:
                        pass
                    try:
                        self.params_dict["before_id"] = int(
                            results[-1]["results"][-1]["id"]
                        )
                        print(
                            f"Retrying and resuming from {int(results[-1]['results'][-1]['id'])}"
                        )
                        self.long_delay *= 2  # double
                        results += self.get_results()
                        break
                    except KeyError:
                        print(
                            "Selected results view does not have 'Reference #' column. Unable to resume download."
                        )
                        raise Exception(
                            "This download is too intense for Formsite servers at the moment. Try using a results view or parameters to only get the results you need."
                        )
                    except Exception as err:
                        print(
                            "This download is too intense for Formsite servers at the moment. Try using a results view or parameters to only get the results you need."
                        )
                        raise err
        return results

    def fetch_results(self, params: dict, page: int) -> dict:
        """Handles fetching and writing (if selected) of results json."""
        self._update_pbar_desc(desc=f"Fetching rows ({(page-1)*500}-{page*500})")
        return self.fetch_content(self.results_url, params, page=page)

    def fetch_content(self, url: str, params: dict, page: int = None) -> dict:
        """Base method for interacting with the formsite api with aiohttp GET request. Returns content of the response."""
        if page is not None:
            params["page"] = page
            if page > 3:
                self._update_pbar_desc(
                    desc=f"Delay [{self.long_delay:0.0f} s] ({(page-1)*500}-{page*500})"
                )
                _ = time.sleep(max(self.long_delay - self.short_delay, 0))
            else:
                self._update_pbar_desc(
                    desc=f"Delay [{self.short_delay:0.0f} s] ({(page-1)*500}-{page*500})"
                )
        time.sleep(self.short_delay)
        with self.session.get(url, params=params) as response:
            content = response.json()
            if response.status_code != 200:
                previous_desc = self.pbar.desc
                self._update_pbar_desc(desc=f"Error: [{response.status_code}]")
                if response.status_code == 429:
                    self._update_pbar_desc(desc="Reached rate limit, waiting 60 seconds")
                    time.sleep(60)
                    self._update_pbar_desc(desc=previous_desc)
                    content = self.fetch_content(url, params, page=page)
                else:
                    err_message = f"[HTTP ERROR {response.status_code}] {response.text} for url '{response.url}'"
                    raise HTTPError(response, err_message)
            if self.check_pages and page is not None:
                self.total_pages = int(response.headers["Pagination-Page-Last"])
                self.check_pages = False
                try:
                    self._update_pbar_total(self.total_pages)
                except AttributeError as ex:
                    # print(ex)
                    pass
        self._update_pbar_progress()
        return content

    def fetch_items(self) -> dict:
        """Handles fetching and writing (if selected) of items json."""
        self._update_pbar_desc(desc="Fetching headers")
        return self.fetch_content(self.items_url, self.items_dict)

    def _update_pbar_total(self, n: int) -> None:
        if isinstance(self.pbar, tqdm):
            self.pbar.total = n

    def _update_pbar_progress(self) -> None:
        if isinstance(self.pbar, tqdm):
            self.pbar.update(1)

    def _update_pbar_desc(self, desc: str) -> None:
        if isinstance(self.pbar, tqdm):
            self.pbar.set_description(desc=desc, refresh=True)



"""This module contains the functionality for downloading downloading many urls concurrently with a worker based queue+semaphore asyncio approach."""

from __future__ import annotations
import os
from collections import namedtuple
import asyncio
import shutil
from typing import Callable, List, Optional, Set
from aiohttp import (
    ClientSession,
    ClientTimeout,
    TCPConnector,
    ClientResponseError,
    InvalidURL,
)

from formsite_util.logger import FormsiteLogger

DownloadItem = namedtuple("DownloadItem", ["url", "path", "attempt"])


class AsyncFormDownloader:

    """Handles the distribution of downlaod tasks across DownloadWorkers."""

    def __init__(
        self,
        download_dir: str,
        sorted_urls: List[str, str],
        workers: int = 5,
        timeout: int = 160,
        max_attempts: int = 1,
        callback: Optional[Callable] = None,
    ) -> None:
        """AsyncFormDownloader constructor

        Args:
            download_dir (str)
            sorted_urls (List[str, str]): List of (url, path)
            workers (int, optional): Number of concurrent downloads. Defaults to 5.
            timeout (int, optional): Download timeout. Defaults to 160.
            max_attempts (int, optional): Max number of attempts. Defaults to 1.
            callback (Optional[Callable], optional): Callback called each time a download is complete. Defaults to None.

        Callback function signature:
            (url: str, path: str, total_files: int) -> None
        """
        # ----
        self.download_dir = download_dir
        self.url_path_list = sorted_urls
        self.workers = workers
        self.timeout = timeout
        self.max_attempts = max_attempts
        self.logger: FormsiteLogger = FormsiteLogger()
        # ----
        self.semaphore = asyncio.Semaphore(self.workers)
        self.dl_queue = asyncio.Queue()
        self.internal_state = DownloadWorkerState(
            self.url_path_list,
            self.workers,
            callback=callback,
        )

    async def run(self) -> None:
        """Entrypoint"""
        os.makedirs(self.download_dir, exist_ok=True)
        async with ClientSession(connector=TCPConnector(limit=0)) as session:
            for url, path in self.url_path_list:
                dl = DownloadItem(url, path, 0)
                self.dl_queue.put_nowait(dl)
            tasks = [
                asyncio.ensure_future(
                    DownloadWorker(
                        self.download_dir,
                        self.dl_queue,
                        self.semaphore,
                        session,
                        self.internal_state,
                        timeout=self.timeout,
                        max_attempts=self.max_attempts,
                    ).run()
                )
                for _ in range(self.workers)
            ]

            await asyncio.gather(*tasks)


class DownloadWorkerState:
    """Contains methods that track internal downlaod progress across downloader workers."""

    def __init__(
        self,
        url_path_list: List[str, str],
        num_workers: int,
        callback: Optional[Callable] = None,
    ) -> None:
        self.url_path_list = url_path_list
        self.num_workers = num_workers
        self.callback = callback
        self.logger: FormsiteLogger = FormsiteLogger()
        # ----
        self.total: int = len(self.url_path_list)
        self.enqueued: int = len(self.url_path_list)
        self.in_progress: int = 0
        self.success: int = 0
        self.failed: int = 0
        self.failed_urls: set = set()
        self.success_urls: set = set()
        self.complete_urls: list = []

    def total_complete(self):
        """Returns sum of `self.success` and `self.failed`"""
        return self.success + self.failed

    def try_exit(self) -> bool:
        """Checks if it is possible to terminate worker."""
        if self.total_complete() >= self.total:
            return True
        elif self.enqueued < self.num_workers:
            return True
        else:
            return False

    def can_terminate_worker(self) -> bool:
        """Decrements allowed number of active workers if conditions allow it."""
        if self.try_exit():
            self.num_workers -= 1
            return True
        else:
            return False

    def mark_success(self, dl: DownloadItem) -> None:
        """Increments internal counter to match completed downloads."""
        self.logger.debug(f"DownloadStatus: Success '{dl.url}' saved in '{dl.path}'")
        self._mark(dl, "OK", self.success_urls, self.success)

    def mark_fail(self, dl: DownloadItem, fail_exception: Exception) -> None:
        """Increments internal counter to match completed downloads."""
        self.logger.debug(f"DownloadStatus: Failure '{dl.url}'")
        self._mark(dl, repr(fail_exception), self.failed_urls, self.failed)

    def _mark(self, dl: DownloadItem, status: str, add_to: Set[str], count_to: int):
        """Base method for changing internal state counts."""
        self.complete_urls.append(f"{dl.url} {status}")
        add_to.add(f"{dl.url}")
        count_to += 1

    def end_iteration(self) -> None:
        """Runs at the end of a download iteration."""
        self.in_progress -= 1

    def start_iteration(self) -> None:
        """Runs at the start of a download iteration."""
        self.enqueued -= 1
        self.in_progress += 1


class DownloadWorker:
    """download_folder is a path to download directory
    queue, semaphore, session, internal state and pbar are shared across all workers"""

    def __init__(
        self,
        download_folder: str,
        queue: asyncio.Queue,
        semaphore: asyncio.Semaphore,
        session: ClientSession,
        internal_state: DownloadWorkerState,
        timeout: int = 160,
        max_attempts: int = 3,
    ) -> None:
        # ----
        self.download_folder = download_folder
        self.queue = queue
        self.semaphore = semaphore
        self.session = session
        self.internal_state = internal_state
        self.timeout = timeout
        self.max_attempts = max_attempts
        self.logger: FormsiteLogger = FormsiteLogger()
        # ----
        self.callback = internal_state.callback
        self.client_timeout = ClientTimeout(total=timeout)

    async def run(self) -> int:
        """Entrypoint for launching the download process."""
        while True:
            try:
                if self.internal_state.can_terminate_worker():
                    break
                await self.semaphore.acquire()
                dl: DownloadItem = await self.queue.get()
                self.internal_state.start_iteration()
                try:
                    await self._fetch(dl.url, dl.path)
                    self.internal_state.mark_success(dl)
                except Exception as download_exception:
                    self._error_handling(download_exception, dl)
            except asyncio.CancelledError:
                pass
            finally:
                self.semaphore.release()
                self.internal_state.end_iteration()

    async def _fetch(
        self,
        url: str,
        path: str,
        chunk_size: int = 4 * 1024,
    ) -> bool:
        """The core download function with `session.get` request."""
        async with self.session.get(url, timeout=self.client_timeout) as response:
            response.raise_for_status()
            with open(f"{path}.tmp", "wb") as f:
                async for chunk in response.content.iter_chunked(chunk_size):
                    f.write(chunk)

        shutil.move(f"{path}.tmp", path)
        if self.callback is not None:
            self.callback(url, path, self.internal_state.total)

    def _error_handling(self, ex: Exception, dl: DownloadItem) -> None:
        """Decides what errors cancel downloads and which are retried."""
        self.logger.debug(f"DownloadError: for '{dl.url}' | {ex}")
        if isinstance(ex, ClientResponseError):
            if ex.status in [403, 404]:
                self.internal_state.mark_fail(dl, ex)
            else:
                self._retry_dl(dl)
        elif isinstance(ex, InvalidURL):
            self.internal_state.mark_fail(dl, ex)
        elif dl.attempt < self.max_attempts:
            self._retry_dl(dl)
        else:
            self.internal_state.mark_fail(dl, ex)

    def _retry_dl(self, dl: DownloadItem) -> None:
        """Puts failed downloads to the end of queue, if attempt < max retry."""
        new_dl = DownloadItem(dl.url, dl.path, dl.attempt + 1)
        self.internal_state.enqueued += 1
        self.logger.debug(f"DownloadStatus: Retry '{dl.url}' attempt {dl.attempt}")
        self.queue.put_nowait(new_dl)



"""
processing.py

This module handles processing of results from API jsons. Returns a dataframe.
"""
from __future__ import annotations
from datetime import datetime, timedelta
from typing import Dict, Iterable
from dataclasses import dataclass
from tqdm import tqdm
import pandas as pd


@dataclass
class _FormsiteProcessing:

    """Handles processing of results from API jsons. Invoked with `self.Process()`

    Args:
        `items` (dict): json of the items API response.
        `results` (List[dict]): list of jsons of the results API response.
        `timezone_offset` (timedelta): offset to shift date_update, date_start, date_finish columns by relative to local timezone. Additive, so to subtract, input negative timedelta.
        `sort_asc` (bool, optional): Whether to sort export by ascending (Reference # or Date if applicable). Defaults to False.
        `use_resultslabels` (bool, optional): True: Use question labels or resultslabels if available. False: Use column IDs and metadata names. Defaults to True.
        `display_progress` (int, optional): If True, displays tqdm progressbar. Defaults to True.
        `params_last` (int, optional): Don't forget to trigger this if you are using the `last` parameter. Trimms excess results. Defaults to None.

    Returns:
        _FormsiteProcessing: An instance of the _FormsiteProcessing class. Start API fetches with `.Start()` method.
    """

    items: dict
    results: Iterable[dict]
    timezone_offset: timedelta
    sort_asc: bool = False
    use_resultslabels: bool = True
    display_progress: bool = True
    params_last: int = None

    def __post_init__(self):
        """Loads json strings as json objects."""
        self.column_map = self._generate_columns()
        self.metadata_map = self._generate_metadata()

    def _generate_columns(self) -> Dict[str, str]:
        """Creates a dict that maps column id to column label from items_json

        Returns:
            Dict[int,str]: Mapping of ID:Label
        """
        column_map = dict()
        try:
            for item in self.items["items"]:
                column_map[str(item["id"])] = item["label"]
        except TypeError:
            pass
        return column_map

    def _generate_metadata(self) -> Dict[str, str]:
        """Creates a dict that maps metadata_label (hardcoded by formsite) to a friendlier export name

        Returns:
            Dict[str,str]: Mapping of metadata_ID: metadata_Label
        """
        return {
            "id": "Reference #",
            "result_status": "Status",
            "date_start": "Start Time",
            "date_finish": "Finish Time",
            "date_update": "Date",
            "user_ip": "User",
            "user_browser": "Browser",
            "user_device": "Device",
            "user_referrer": "Referrer",
            "user_os": "OS",
            "payment_status": "Payment Status",
            "payment_amount": "Payment Amount Paid",
            "login_username": "Login Username",
            "login_email": "Login Email",
            "score": "Score",
        }

    def _process_items_row(self, items: dict) -> pd.Series:
        """Converts a single 'items' record to a pandas Series

        Args:
            items (dict): ['items'] key from a results dict record

        Returns:
            pd.Series: a row of the new output csv
        """
        cols = []
        row = []
        for t in items:
            cols.append(str(t["id"]))
            if t.get("values") is not None:
                values = []
                for val in t["values"]:
                    values.append(val["value"])
                value = " | ".join(values)
            else:
                value = t["value"]
            row.append(value)
        return pd.Series(row, cols)

    def _process_metadata_row(self, result: dict) -> pd.Series:
        """Converts all non-['items'] keys to a pd.Series row

        Args:
            result (dict): an single results json record

        Returns:
            pd.Series: A row containing all metadata-related information
        """
        if "items" in result:
            del result["items"]
        return pd.Series(result)

    def _process_row(self, in_json: dict):
        """Merges metadata_row and items_row into a single row.

        Args:
            in_json (dict): a single results record json

        Returns:
            pd.Series: a row
        """
        items = self._process_items_row(in_json["items"])
        metadata = self._process_metadata_row(in_json)
        concated = pd.concat((metadata, items), axis=0)
        return concated

    def _reorder_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        """Orders existing columns into a standard export format.
        First is always Reference #, Status, then Items, then Remaining metadata.

        Args:
            df (pd.DataFrame): Unordered DataFrame

        Returns:
            pd.DataFrame: Ordered DataFrame
        """
        left_side = []
        right_side = []
        if "id" in df.columns:
            left_side.append("id")
        if "result_status" in df.columns:
            left_side.append("result_status")

        if len(self.column_map.keys()) > 0:
            middle = [
                col
                if (col in df.columns) and (col not in self.metadata_map.keys())
                else None
                for col in list(self.column_map.keys())
            ]
        else:
            middle = []
            for col in list(df.columns):
                if col not in list(self.metadata_map.keys()):
                    middle.append(col)

        while None in middle:
            middle.remove(None)
        if "payment_status" in df.columns:
            right_side.append("payment_status")
        if "payment_amount" in df.columns:
            right_side.append("payment_amount")
        if "login_username" in df.columns:
            right_side.append("login_username")
        if "login_email" in df.columns:
            right_side.append("login_email")
        if "score" in df.columns:
            right_side.append("score")
        if "date_update" in df.columns:
            right_side.append("date_update")
        if "date_start" in df.columns:
            right_side.append("date_start")
        if "date_finish" in df.columns:
            right_side.append("date_finish")
        if "Duration (s)" in df.columns:
            right_side.append("Duration (s)")
        if "user_ip" in df.columns:
            right_side.append("user_ip")
        if "user_browser" in df.columns:
            right_side.append("user_browser")
        if "user_device" in df.columns:
            right_side.append("user_device")
        if "user_referrer" in df.columns:
            right_side.append("user_referrer")

        final = left_side + middle + right_side
        return df[final]

    def _cast_dtypes(self, df: pd.DataFrame) -> pd.DataFrame:
        """Casts the type of each existing column to their standard form.

        Args:
            df (pd.DataFrame): DataFrame with erratic dtypes

        Returns:
            pd.DataFrame: DataFrame with set dtypes
        """
        if "id" in df.columns:
            df["id"] = df["id"].astype(int, errors="ignore")
        if "result_status" in df.columns:
            df["result_status"] = df["result_status"].astype(str)
        if "payment_status" in df.columns:
            df["payment_status"] = df["payment_status"].astype(str)
        if "payment_amount" in df.columns:
            df["payment_amount"] = df["payment_amount"].astype(str)
        if "login_username" in df.columns:
            df["login_username"] = df["login_username"].astype(str)
        if "score" in df.columns:
            df["score"] = df["score"].astype(int, errors="ignore")
        if "login_email" in df.columns:
            df["login_email"] = df["login_email"].astype(str)
        if "date_update" in df.columns:
            df["date_update"] = df["date_update"].apply(self._string2datetime)
        if "date_start" in df.columns:
            df["date_start"] = df["date_start"].apply(self._string2datetime)
        if "date_finish" in df.columns:
            df["date_finish"] = df["date_finish"].apply(self._string2datetime)
        if "Duration (s)" in df.columns:
            df["Duration (s)"] = df["Start Time"] - df["Start Time"]
            df["Duration (s)"] = df["Duration (s)"].apply(lambda x: x.total_seconds())
        if "user_ip" in df.columns:
            df["user_ip"] = df["user_ip"].astype(str)
        if "user_browser" in df.columns:
            df["user_browser"] = df["user_browser"].astype(str)
        if "user_device" in df.columns:
            df["user_device"] = df["user_device"].astype(str)
        if "user_referrer" in df.columns:
            df["user_referrer"] = df["user_referrer"].astype(str)
        return df

    def _sort_data(self, df: pd.DataFrame, ascending_bool: bool) -> pd.DataFrame:
        try:
            df = df.sort_values(by=["Reference #"], ascending=ascending_bool)
        except KeyError:
            try:
                df = df.sort_values(by=["Date"], ascending=ascending_bool)
            except KeyError:
                pass

        return df.reset_index(drop=True)

    def Process(self) -> pd.DataFrame:
        """Loads jsons in results list as dataframes and concats them."""
        results_list = []
        for results in [json["results"] for json in self.results]:
            results_list += results

        series_list = [
            self._process_row(json_row)
            for json_row in tqdm(
                results_list,
                desc="Processing results",
                unit=" rows",
                ncols=80,
                dynamic_ncols=True,
                leave=False,
            )
        ]

        dataframe = pd.DataFrame(series_list)
        dataframe = self._reorder_columns(dataframe)
        dataframe = self._cast_dtypes(dataframe)
        if self.use_resultslabels:
            combined_map = self.column_map
            combined_map.update(self.metadata_map)
            dataframe.rename(columns=combined_map, inplace=True)
        if self.params_last is not None:
            dataframe = self._sort_data(dataframe, False)
            until = int(self.params_last)
            dataframe = dataframe.head(until)
        dataframe = self._sort_data(dataframe, self.sort_asc)
        return dataframe

    def _string2datetime(self, old_date: str) -> datetime:
        """Converts ISO 8601 datetime string to datetime class."""
        try:
            new_date = datetime.strptime(
                old_date, "%Y-%m-%dT%H:%M:%SZ"
            )  # ISO 8601 standard
            new_date = new_date + self.timezone_offset
            return new_date
        except TypeError:
            return old_date



"""Defines FormsiteForm object and its logic."""

from __future__ import annotations
import os
from pathlib import Path
from time import sleep
import re
from typing import Callable, Generator, List, Optional
import pandas as pd
from requests import Session

# ----
from formsite_util.error import FormsiteNoResultsException
from formsite_util.parameters import FormsiteParameters
from formsite_util.form_fetcher import FormFetcher
from formsite_util.form_parser import FormParser
from formsite_util.download import DownloadStatus, download_sync, filter_urls
from formsite_util.download_async import AsyncFormDownloader
from formsite_util.form_data import FormData
from formsite_util.cache import (
    items_load,
    items_match_data,
    items_save,
    results_load,
    results_save,
)


def tz_shif_inplace(df: pd.DataFrame, col: str, tz_name: str):
    """Converts a tz-aware dataframe column to target timezone"""
    if col in df:
        df[col] = df[col].dt.tz_convert(tz=tz_name)


class FormsiteForm(FormData):
    """Formsite API Form object, representing the data and HTTP session"""

    def __init__(
        self,
        form_id: str,
        token: str,
        server: str,
        directory: str,
        prepoulate_data: FormData = None,
    ):
        """FormsiteForm constructor

        Args:
            form_id (str): Formsite Form ID
            token (str): Formsite API Token
            server (str): Formsite Server (fsX.formsite.com)
            directory (str): Formsite Directory
            prepoulate_data (FormData, optional): Prepopulate form with this FormData object
        """

        super().__init__()
        self.form_id: str = form_id
        self.token: str = token
        self.server: server = server
        self.directory: directory = directory
        if prepoulate_data is not None:
            self._data = prepoulate_data._data
            self._items = prepoulate_data._items
            self._labels = prepoulate_data._labels

        self.logger.debug(f"Initialized {repr(self)}")

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.form_id}>"

    def fetch(
        self,
        fetch_results: bool = True,
        fetch_items: bool = True,
        params: FormsiteParameters = FormsiteParameters(),
        result_labels_id: int = None,
        fetch_delay: float = 3.0,
        fetch_callback: Callable = None,
        # ---
        cache_items_path: str = None,
        cache_results_path: str = None,
    ):
        """Updates the FormsiteForm object instance with the data from Formsite accoridng to input params

        Fetch Args:
            fetch_results (bool, optional): Pull results within parameters. Defaults to True.
            fetch_items (bool, optional): Pull items (of resultslabels id). Defaults to True.
            params (FormsiteParameters): Pull results and items according to these parameters.
            result_labels_id (int, optional): Results labels id to use when fetching items. Defaults to None.
            fetch_delay (float): Time delay between individual API calls in seconds.
            fetch_callback (Callable, optional): Run this callback every time an API fetch is complete.

        Cache Args:
            cache_items_path (str) : Path where to store form items (as json). Defaults to None (will not store).
            cache_results_path (str) : Path where to store form results. Defaults to None (will not store).


        Supported Cache formats (file extensions) are:
            - parquet (recommended)
            - pkl | pickle
            - feather
            - hdf

        Callback funciton signature:
            function(cur_page: int, total_pages: int, data: dict) -> None
        """
        params = params.copy()
        self.logger.debug(f"{repr(self)} fetching with {params}")
        # -!- RESULTS PART
        parser = FormParser()
        fetcher = FormFetcher(
            self.form_id,
            self.token,
            self.server,
            self.directory,
            params,
        )
        if fetch_results:
            # ---- handle cache ----
            if cache_results_path is not None:
                if not isinstance(cache_results_path, str):
                    raise TypeError("Invalid path")
                cached_results = results_load(cache_results_path)
                if isinstance(cached_results, pd.DataFrame) and not cached_results.empty:
                    if "id" not in cached_results.columns:
                        raise ValueError(
                            "Expected stored data to have the 'id' (Reference #) column. Add this to your results view."
                        )
                    self.logger.debug(
                        f"Cache results {self.form_id}: Overwriting after_id:{max(cached_results['id'])} | before_id:None"
                    )
                    params.after_id = max(cached_results["id"])
                    params.before_id = None
                    fetcher.params = params
            # -!!- perform results fetch -!!-
            for data in fetcher.fetch_iterator():
                # --- edge case ---
                if not data.get("results") and not (
                    isinstance(cached_results, pd.DataFrame) and not cached_results.empty
                ):
                    raise FormsiteNoResultsException("No results in specified parameters")
                # --- regular case ---
                parser.feed(data)
                # --- callback ---
                if isinstance(fetch_callback, Callable):
                    fetch_callback(fetcher.cur_page, fetcher.total_pages, data)
                # --- fetch delay ---
                sleep(fetch_delay)
            # ---- finish handling cache ----
            if cache_results_path is not None:
                new_data = parser.as_dataframe()
                self.logger.debug(
                    f"Cache results {self.form_id}: Appending {new_data.shape[0]} new results"
                )
                # --- if there are new results, merge ---
                if new_data.shape[0] > 0:
                    merged_results = pd.concat(
                        [new_data, cached_results], ignore_index=True
                    )
                    merged_results = merged_results.reset_index(drop=True)
                    self.data = merged_results
                    results_save(merged_results, cache_results_path)
                # --- otherwise just use the data we got ---
                else:
                    self.data = cached_results
            else:
                self.data = parser.as_dataframe()
            tz_shif_inplace(self.data, "date_update", params.timezone)
            tz_shif_inplace(self.data, "date_start", params.timezone)
            tz_shif_inplace(self.data, "date_finish", params.timezone)
            if params.last is not None:
                self.data = self.data.head(params.last)

        # -!- ITEMS PART
        if fetch_items:
            if cache_items_path is not None:
                if not isinstance(cache_items_path, str):
                    raise TypeError("Invalid path")
                cached_results = items_load(cache_items_path)
                if cached_results is None or not items_match_data(
                    cached_results, self.data.columns
                ):
                    self.logger.debug(
                        f"Cache items {self.form_id}: Fetching new items for cache"
                    )
                    cached_results = fetcher.fetch_items(result_labels_id)
                    items_save(cached_results, cache_items_path)
                self.items = cached_results
            else:
                self.items = fetcher.fetch_items(result_labels_id)
            self.labels = parser.create_rename_map(self.items)

    def downloader(
        self,
        download_dir: str,
        timeout: float = 160,
        max_attempts: int = 3,
        url_filter_re: str = r".+",
        filename_substitution_re_pat: str = r"",
        strip_prefix: bool = False,
        overwrite_existing: bool = True,
    ) -> Generator[DownloadStatus, None, None]:
        """Download files uploaded to the form via the File upload control (sequential)

        Generator yields:
            4-tuple of (url: str, filename: str, path: str, status: DownloadStatus)

        Args:
            download_dir (str): Directory to download files into. Gets created if it doesn't exist.
            timeout (float): Seconds to wait for the file to download. Defaults to 160.
            max_attempts (int): How many times to retry the download in case of failure. Defaults to 3.
            url_filter_re (str): Keep only URLs that match this regex. Defaults to r".+".
            filename_substitution_re_pat (str): Remove all charactes that match the regex. Defaults to r"".
            strip_prefix (bool): Strip the f-xxx-xxx-  or sig-xxx-xxx- prefix from filenames. Defaults to False.
            overwrite_existing (bool): Check download directory for files. If they already exist, don't download them. Defaults to True.
        """
        download_dir = Path(download_dir).resolve().as_posix()
        os.makedirs(download_dir, exist_ok=True)
        urls = self.extract_urls(url_filter_re)
        filtered_URLs = filter_urls(
            urls,
            download_dir,
            strip_prefix=strip_prefix,
            filename_substitution_re_pat=filename_substitution_re_pat,
            overwrite_existing=overwrite_existing,
        )
        # ----
        with Session() as session:
            for url, path in filtered_URLs:
                status = download_sync(
                    url,
                    path,
                    session,
                    timeout=timeout,
                    max_attempts=max_attempts,
                )
                yield status

    def async_downloader(
        self,
        download_dir: str,
        max_concurrent: int = 5,
        timeout: float = 160,
        max_attempts: int = 3,
        url_filter_re: str = r".+",
        filename_substitution_re_pat: str = r"",
        strip_prefix: bool = False,
        overwrite_existing: bool = True,
        callback: Optional[Callable] = None,
    ) -> AsyncFormDownloader:
        """Download files uploaded to the form via the File upload control (async)

        Returns a coroutine that runs the download process.

        Args:
            download_dir (str): Directory to download files into. Gets created if it doesn't exist.
            max_concurrent (int): Maximum concurrent downloads allowed. Defaults to 5.
            timeout (float): Seconds to wait for the file to download. Defaults to 160.
            max_attempts (int): How many times to retry the download in case of failure. Defaults to 3.
            url_filter_re (str): Keep only URLs that match this regex. Defaults to r".+".
            filename_substitution_re_pat (str): Remove all charactes that match the regex. Defaults to r"".
            strip_prefix (bool): Strip the f-xxx-xxx-  or sig-xxx-xxx- prefix from filenames. Defaults to False.
            overwrite_existing (bool): Check download directory for files. If they already exist, don't download them. Defaults to True.
            callback (Optional[Callable], optional): Callback called each time a download is complete. Defaults to None.
        """

        download_dir = Path(download_dir).resolve().as_posix()
        os.makedirs(download_dir, exist_ok=True)
        urls = self.extract_urls(url_filter_re)
        filtered_URLs = filter_urls(
            urls,
            download_dir,
            strip_prefix=strip_prefix,
            filename_substitution_re_pat=filename_substitution_re_pat,
            overwrite_existing=overwrite_existing,
        )
        # ----
        downloader = AsyncFormDownloader(
            download_dir,
            filtered_URLs,
            max_concurrent,
            timeout,
            max_attempts,
            callback,
        )
        # ----
        return downloader

    def extract_urls(self, filter_re_pat=r".+") -> List[str]:
        """Extract all URLs of files uploaded to the form

        Args:
            filter_re_pat (regexp, optional): Output only the URLs that match the input regex. Defaults to r".+".

        Returns:
            List[str]: List of URLs to files uploaded to the form.
        """
        url_re_pat = (
            rf"(https\:\/\/{self.server}\.formsite\.com\/{self.directory}\/files\/.*)"
        )
        url_re = re.compile(url_re_pat)
        urls = set()
        for col in self.data.columns:
            try:
                url_mask: pd.Index = self.data[col].str.fullmatch(url_re) == True
                tmp: pd.Series = self.data[url_mask][col]
                tmp = tmp.str.split("|")
                tmp = tmp.explode().str.strip()
                urls = urls.union(tmp.to_list())
            except AttributeError:
                pass

        # Return all URLs that match filter_re_pat
        filter_re = re.compile(filter_re_pat)
        return sorted([url for url in urls if filter_re.match(url)])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['formsite_util', 'formsite_util.legacy']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'colorama>=0.4.4,<0.5.0',
 'pandas>=1.3,<2.0',
 'pytz>=2021.3,<2022.0',
 'requests>=2.27.1,<3.0.0',
 'tqdm>=4.62.3,<5.0.0']

entry_points = \
{'console_scripts': ['getform = formsite_util.cli:main']}

setup_kwargs = {
    'name': 'formsite-util',
    'version': '2.0.2',
    'description': 'Python library and CLI tool for interacing with the FormSite API',
    'long_description': '<!-- HEADER -->\n<br />\n<p align="center">\n  <a>\n    <img src="https://www.formsite.com/wp-content/themes/formsite-theme/assets/favicon/favicon-96x96.png" alt="Logo" width="80" height="80">\n  </a>\n\n  <h3 align="center">Formsite Utility</h3>\n</p>\n\nPython library and CLI tool for interacing with the FormSite API\n\n[![Codacy Badge](https://app.codacy.com/project/badge/Grade/31145f1b97934080981a53783803701f)](https://www.codacy.com/gh/strny0/formsite-utility/dashboard?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=strny0/formsite-utility&amp;utm_campaign=Badge_Grade)\n[![PyPI version](https://badge.fury.io/py/formsite-util.svg)](https://badge.fury.io/py/formsite-util)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/formsite-util)\n\n## Quickstart\n\n### Install\n\n```bash\n$ pip install formsite-util\n```\n\n### CLI\n\nTOKEN: Formsite API access token, acquired from account settings\n\nLook at the formsite URL when accessing one of your forms\n\nhttps:/<span></span>/`SERVER`.formsite.com/`DIRECTORY`/`FORM_ID`\n\nExport a form:\n\n```bash\n$ getform -t TOKEN -s SERVER -d DIRECTORY -f FORM_ID -o ./my_export.csv\n```\n\nDownload files uploaded to a form:\n\n```bash\n$ getform -t TOKEN -s SERVER -d DIRECTORY -f FORM_ID -D ./download_dir/\n```\n\n### Module\n\n```python\nfrom formsite_util import FormsiteForm, FormsiteFormsList\n\n\n# Formsite Form\nform = FormsiteForm(FORM_ID, TOKEN, SERVER, DIRECTORY)\nform.fetch()\n\ndf = form.data # data with columns as column IDs\ndf = form.data_labels # data with columns as the actual labels\n\n# work with df\n...\n\n# Formsite Forms List\nformlist = FormsiteFormsList(TOKEN, SERVER, DIRECTORY)\nformlist.fetch()\n\nforms = formlist.data\n# work with forms\n...\n```\n\n## Overview\n\nThis program performs an export of a specified formsite form with parameters\nA faster alternative to a manual export from the formsite website. It uses the formsite API v2 to perform exports of results. You can specify parametrs in commandline to filter these results. Allow advanced features like link extraction and even download of files to a specified directory.\n\nSupported python versions: | `3.8` | `3.9` | `3.10+` |\n\n### Dependencies\n\n```txt\naiohttp\nrequests\ntqdm\npytz\npandas\npyarrow # parquet and feather support\nopenpyxl # excel support\ntables # hdf support\n```\n\n## Usage\n\nYou can invoke the module with `getform [args]` (if it\'s in your [PATH](https://datatofish.com/add-python-to-windows-path/))\n\nor `py -m formsite_util.cli`\n\nor `python3 -m formsite_util.cli`\n\nYou can access the help page with `getform -h`\n\n## **CLI Documentation:**\n\n### **Authorization arguments:**\n\nAuthorization arguments are required for all operations.\n\n```bash\ngetform -t \'TOKEN\' -s \'SERVER\' -d \'DIRECTORY\'\n```\n\n**Token**: Your Formsite API token.\n\n**Server**: Your Formsite server. A part of the url. https:/<span></span>/`fsX`.forms… <- the \'fsX\' part. For example \'fs4\'.\n\n**Directory**: Can be found under [Share > Links > Directory] It is the highlighted part of the url: https:/<span></span>/fsXX.formsite.com/`directory`/hfjfdyjyf\n\n### **Results parameters:**\n\nResults parameters are used to set filters upon the results you are retreiving from a specific form.\n\n**Form**: Can be found under [Share > Links > Directory] It is the highlighted part of the url: https:/<span></span>/fsXX.formsite.com/gnosdH/`form_id` or by running `getform -l` to list all forms and their IDs\n\n#### **Filter by Reference #**\n\nYou can provide arguments `--afterref *int* and --beforeref *int*` to specify an interval of which reference numbers to get from your export. It is the same as setting a filter based on reference number when doing an export from the formsite website.\n\nExample:\n\n```bash\n$ getform -t \'token\' -d \'directory\' -s \'server\' -f \'form_id\'  \\\n--afterref 14856178 --beforeref 15063325 \\\n-o\n```\n\nThis will output results in between reference numbers `14856178` and `15063325`.\n\n#### **Filter by Date**\n\nYou can provide arguments `--afterdate *date* and --beforedate *date*` to specify an interval of which dates to get from your export. It is the same as setting a filter based on date number when doing an export from the formsite website.\n\nValid datetime formats are `ISO 8601`, `yyyy-mm-dd` or `yyyy-mm-dd HH:MM:SS`\n\n```txt\nISO 8601 example\nyyyy-mm-ddTHH:MM:SSZ\n2021-01-20T12:00:00Z is 20th January, 2021 at noon in UTC timezone\n```\n\nExample:\n\n```bash\n$ getform -t \'token\' -d \'directory\' -s \'server\' -f \'form_id\'  \\\n--afterdate \'2021-01-01T00:00:00Z\' ---beforedate \'2021-02-01T00:00:00Z\' \\\n-o\n```\n\nThis will retrieve all results in for the month of January 2021 in UTC.\n\n#### **Timezone:**\n\nThe dates provided are in UTC. This can become a problem if your organizations formsite account is set to a particular timezone you are not in, especially when deleting results/attachments using date filters.\n\nYou can set a manual offset with the `-T` or `--timezone` arguments.\nValid input is a timezone database name such as `America/Chicago`. This will shift formsite statistics dates *(columns Date | Start time | Finish time)* and your before/after date argument to your input timezone to the same UTC time but in the target timezone.\n\nExample usage: I want to get results for a certain day in non-UTC timezone. Passing `-T \'America/Chicago\'`, my input args `--afterdate 2021-04-10 --beforedate 2021-4-11` will become `2021-04-09 18:00:00` and `2021-04-10 18:00:00`. Additionally, columns Date, Start time, Finish time will also be shifted.\n\nExample:\n\n```bash\n$ getform -t \'token\' -d \'directory\' -s \'server\' -f \'form_id\' --afterdate \'2021-02-01T00:00:00Z\' --beforedate \'2021-02-01T06:30:00Z\' -T \'America/New_York\' -o\n```\n\n#### **Sorting:**\n\nInvoked with `--sort` [asc|desc] ex: `--sort desc`\n\nSorts rows based on Reference # column. Defaults to descending.\n\n#### **Results labels and views:**\n\nResults labels are the names of columns of items you put on your form. The default option is to use the default question labels you would put on your form, such as "What is your email address?" or "Your email:" or simply "E-mail:" but for actually working with the data, it is easier and clearer to use a feature formsite supports called Results labels, where you can set your own names of these columns for exports.\n\nYou can find the ID of your labels under `[Form Settings > Integrations > Formsite API]`\n\nYou can set them with the `--resultslabels *id*` argument. Defaults to 10 which are typically the first results label you add.\n\nYou can also set results view with `--resultsview *id*` argument. Defaults to 11 which is all items + statistics.\n\n### **Outputing to a file:**\n\nYou can use the `-o` flag to output your export to a file. If you don\'t specify this flag, no results will be outputted. For reasons as to why you wouldn\'t include this flag, please see ***File downloads:*** below.\n\nYou can leave the `-o` flag by itself or you can specify a path, either relative or absolute to a file you want to output to. If you don\'t include a path, it will default to its current directory, the file will be a csv with the name in the following format:\n\n```txt\nexport_formID_date.csv | example: export_formID_2021-04-08--20-00-18.csv\n```\n\nThe output file changes completely based on what extension you give it. Supported filetypes are:\n\n | `.csv`\n | `.xlsx`\n | `.parquet`\n | `.feather`\n | `.pkl | .pickle`\n | `.hdf`\n\nExample:\n\n```bash\n$ getform -t \'token\' -d \'directory\' -s \'server\' -f \'form_id\'  \\\n-o \'./exports/export_filename.csv\'\n```\n\n```bash\n$ getform -t \'token\' -d \'directory\' -s \'server\' -f \'form_id\'  \\\n-o \'./output.xlsx\'\n```\n\n#### **CSV Encoding:**\n\nSpecify encoding of the output file (if output format supports it). Defaults to \'utf-8-sig\'.\n\nInvoked with `--encoding utf-8`\n\n#### **CSV Delimiter/Separator:**\n\nSpecify separator of the output file (if output format supports it). Defaults to \',\' (comma)\n\nInvoked with `--separator \',\'`\n\n#### **CSV Line ending:**\n\nSpecify line terminator of the output file (if output format supports it).\n\nCan be one of: { LF, CR, CRLF, os_default }\n\nInvoked with: `--line_terminator os_default`\n\n#### **CSV Quoting:**\n\nSpecify quoting level of the output file (if output format supports it). Defaults to \'MINIMAL\'\n\nMore info about the quoting levels: <https://docs.python.org/3/library/csv.html>\n\nCan be one of: { QUOTE_ALL, QUOTE_MINIMAL, QUOTE_NONNUMERIC, QUOTE_NONE }\n\nInvoked with: `--quoting QUOTE_MINIMAL`\n\n#### **Date format:**\n\nInvoked with `--date_format format` where format is a string of python datetime directives.\n\nDefaults to `\'%Y-%m-%d %H:%M:%S\'` which is `yyyy-mm-dd HH:MM:SS`\n\nYou can find the possible format directives here: <https://docs.python.org/3/library/datetime.html#strftime-and-strptime-behavior>\n\nSets the date format of all formsite statistics dates such as the Date column. Only applies to CSV output files as excel handles it by itself.\n\n### **File downloads:**\n\nYou can use the `-x` flag to extract all links that begin with the formsite base url into a text file, such as:\n\n```txt\nhttps://fsXX.formsite.com/directory/files/f-XXX-XXX-reference#_filename.ext\n```\n\nThe links file defaults to the same directory as the `getform.py` with the file name `links_{formID}_{timestamp}.txt`\n\nYou can use the `-D` option to download all links to a directory you specify, just like with the `-o` argument. The file will be saved with the filename in the link `reference#_filename.ext`\n\nInvoked with: `-D` or `--download` \'path/to/folder\' | `-x` or `--extract` path/to/file\n\nExample:\n\n```bash\n$ getform -t \'token\' -d \'directory\' -s \'server\' -f \'form_id\'  \\\n-D \'./download_03/\' -xre \'\\.jpg$\'\n```\n\n*^Example (will create a directory download_03 in the folder where you run it and save all files that end with .jpg uploaded to the form)*\n\n#### **Links regex:**\n\nYou can also specify a regex to filter links. You will only get links that contain a match of a regex you provide with the `-xre` or `--extract_regex` option:\n\n```txt\nex: \\.json$   - will only return files that end with .json\n```\n\n#### **Max concurrent downloads:**\n\nYou can specify how many concurrent downloads can take place at any given moment.\n\nInvoked with: `-c x` or `--concurrent_downloads x`\n\nDefaults to `10`\n\n#### **Stop file overwrite:**\n\nIf you include this flag, files with the same filenames as you are downloading in the target folder will not be overwritten and re-downloaded.\n\nInvoked with: `-n` or `--dont_overwrite_downloads`\n\n#### **Timeout:**\n\nTimeout in seconds for each individual file download.\n\nIf download time exceeds it, it will throw a timeout error and retry up until retries.\n\nInvoked with: `--timeout x`\n\nDefaults to `160` seconds\n\n#### **Retries:**\n\nNumber of times to retry downloading files if the download fails.\n\nInvoked with: `--retries x`\n\nDefaults to `3`\n\n#### **Strip prefix:**\n\nIf you enable this option, filenames of downloaded files will not have f-xxx-xxx- prefix.\n\nInvoked with: `--strip_prefix`\n\n#### **Filename regex:**\n\nIf you include this argument, filenames of the files you download from formsite servers will remove characters that dont match the regex from their filename.\n\nInvoked with: `-Dre` or `--download_regex`\n\nExample: `-Dre \'[^\\w\\_\\-]+\'` will only keep alphanumeric chars, underscores and dashes in filename.\n\n*^(in case of filename colissions, appends _number)*\n\n### **Other arguments:**\n\n#### **Special options:**\n\n`-h --help` - shows a help message and exits\n\n`--disable_progressbars` or `-P` If you use this flag, program will not display progressbars to console\n\n`-l --list_forms` - prints all forms, can be saved as csv if you provide a path to a file (`-l list.csv`).\n\n*^You can pair this with `getform (...) -l | grep form name` to find form ID easily.*\n\n*Or pipe it into less `getform (...) -l | less` to browse it*\n\n\n`-V --version` - displays the current version and exits\n\n`-v --verbose` - displays logger *(level DEBUG)* information to stdout, disables progress bars\n\n## **Module Examples:**\n\n### formsite_util\n\nThe formsite-util package provides several interfaces for common tasks.\n\n### High level interfaces\n\nFormsiteForm: Represents the form data and session\n\nFormsiteFormsList: Represents the list of all forms for the specified account\n\nFormsiteParameters: Represents parameters for results/items requests\n\n### Low level interfaces\n\nFormFetcher: Result/Item fetching operations *(useful)*\n\nFormParser: Result/Item parsing operations *(useful)*\n\nFormData: Represents the base form data class *(not very useful)*\n\nFormsiteLogger: Custom logger you may connect to your own logging *(not very useful)*\n\n## Module example usage\n\n### FormsiteForm example usage\n\n```python\nfrom formsite_util import FormsiteForm, FormsiteParameters\n\ntoken = "efwfjwi0fj0W4JG340G343G" # not real token\nserver = "fs1"\ndirectory = "aqWfcw"\n\n# Get all data for the month of January 2022\nparams = FormsiteParameters(\n    after_date="2022-01-01T00:00:00Z", \n    before_date="2021-02-01T00:00:00Z",\n    )\n\nform = FormsiteForm(form_id, token, server, directory)\nform.fetch(params=params)\nform.data_labels.to_csv(\'./2022_jan.csv\')\n```\n\n### FormsiteFormsList example usage\n```python\nfrom formsite_util import FormsiteFormsList, FormsiteForm\n\n# Iterate through all forms for the account and fetch them\nflist = FormsiteFormsList(token, server, directory)\nflist.fetch()\n\nfor form_id in flist.data[\'form_id\'].to_list():\n    form = FormsiteForm(form_id, token, server, directory)\n    form.fetch()\n    form.to_csv(f\'./expors/{form_id}_data.csv\')\n```\n\n### Caching example\n\nWhen fetching a form, it is wasteful to fetch everything again.\n\nWe wish to spare the API calls due to the rate limit.\n\nFor this reason, you may use caching of the items and results.\n\n```python\nfrom formsite_util import FormsiteForm\n\nform = FormsiteForm(form_id, token, server, directory)\n\n\n# By adding these parameters to the fetch request...\nform.fetch(\n    cache_items_path=f\'./cache_dir/{form_id}_items.json\',\n    cache_results_path=f\'./cache_dir/{form_id}_data.parquet\',\n)\n\n# For Items:\n#   - You are only re-fetching them if the results labels column IDs don\'t match the results column IDs (ie. only when necessary)\n\n\n# For Results:\n#   - You are only fetching results since the latest \'id\' (aka Reference #) and merging them back\n```\n\n### Connecting logging\n\n```python\nimport logging\n\n# Create your desired handler object\n# Example Stream handler\nhandler = logging.StreamHandler(sys.stdout)\nfmt = logging.Formatter("%(message)s")\nhandler.setLevel(logging.DEBUG)\nhandler.setFormatter(fmt)\n\n# Add it to the form.logger object using the addHandler method\nform.logger.addHandler(handler)\n\n# NOTE: You only need to do this once for all form objects.\n#       FormsiteLogger is a singleton, only 1 instance may exist at any time.\n#       To do this once at the start of your program, you can do this:\n\nfrom formsite_util.logger import FormsiteLogger\nl = FormsiteLogger()\nl.addHandler(...)\n\n```\n\n### FormsiteForm class\n\n```python\nfrom formsite_util import FormsiteForm, FormsiteParameters\n\nform = FormsiteForm(...)\nparams = FormsiteParameters(...)\n\nform.fetch(fetch_results=True, # Populates form.data\n           fetch_items=True,   # Populates form.items, form.labels\n           fetch_delay=3.0,    # delay between each API call in seconds\n           params=params,      # Results parameters\n           result_labels_id=0, # Items labels ID found in [Form settings -> Integrations -> Formsite API]\n           )\n\ndf1 = form.data \n# DataFrame with columns as column IDs used by formsite\n# returns data if results were fetched\n\ndf2 = form.data_labels \n# DataFrame with columns as labels of (result_labels_id)\n# data_labels only returns data, if both results and items were fetched\n\nlabels_dict = form.labels \n# The mapping of column ID -> label\n# returns data if items were fetched\n\nitems_dict = form.items \n# The original, unparsed formsite items object\n# returns data if items were fetched\n\n\nform.extract_urls(...)\n# Extracts URLs to all files uploaded to the form using bult in controls\n\nform.downloader(...)\n# Launches a syncrhonous downloader of the files uploaded to the form\n\nform.async_downloader(...)\n# Launches an async downloader of the files uplaoded to the form\n\nform.to_csv(...)\n# Saves the form data or data_labels to a CSV file with the following default settings:\n#   - by default save data_labels\n#   - encoded as utf-8-sig (utf-8 with BOM)\n#   - index=False\n#   - %Y-%m-%d %H:%M:%S datetime format\n\nform.to_excel(...)\n# Saves the form data or data_labels to an Excel file with the following default settings:\n#   - by default save with data_labels\n#   - index=False\n\n```\n\n### FormsiteFormsList class\n\n```python\nfrom formsite_util import FormsiteFormsList\n\nflist = FormsiteFormsList(...)\n\nflist.fetch()\n\nflist.data\n# DataFrame in the format:\n# form_id | name | state | results_count | files_size | files_size_human | url\n# ...\n```\n\n## Notes\n\nMore info can be found at Formsite API v2 help page\n\n**<https://support.formsite.com/hc/en-us/articles/360000288594-API>**\n\nYou can find API related information of your specific form under:\n\n```txt\n[Form Settings > Integrations > Formsite API] on the Formstie website\n```\n\n**API response error codes table:**\n\n| code | description                                 |\n|------|---------------------------------------------|\n| 401  | Authentication info is missing or invalid.  |\n| 403  | Forbidden.                                  |\n| 404  | Path or object not found.                   |\n| 422  | Invalid parameter.                          |\n| 429  | Too many requests or too busy.              |\n| 5xx  | Unexpected internal error.                  |\n\n## License\n\n© 2022 Jakub Strnad\n\nMIT License\nPlease see LICENSE.<span></span>md for more details.\n',
    'author': 'Jakub Strnad',
    'author_email': 'jakub.strnad@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/strny0/formsite-utility/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Stores useful constants."""
from csv import QUOTE_ALL, QUOTE_MINIMAL, QUOTE_NONE, QUOTE_NONNUMERIC
from datetime import datetime as dt
from os import linesep

# Columns that appear outside of the "items" object in the response json
METADATA_COLS = {
    "id": "Reference #",
    "result_status": "Status",
    "login_username": "Username",
    "login_email": "Email Address",
    "payment_status": "Payment Status",
    "payment_amount": "Payment Amount Paid",
    "score": "Score",
    "date_update": "Date",
    "date_start": "Start Time",
    "date_finish": "Finish Time",
    "user_ip": "User",
    "user_browser": "Browser",
    "user_device": "Device",
    # "user_os": "OS", # Deprecated, not present in exports
    "user_referrer": "Referrer",
}

HTTP_429_WAIT_DELAY = 60  # seconds
ConnectionError_DELAY = 10  # seconds

QUOTE = {
    "QUOTE_ALL": QUOTE_ALL,
    "QUOTE_MINIMAL": QUOTE_MINIMAL,
    "QUOTE_NONE": QUOTE_NONE,
    "QUOTE_NONNUMERIC": QUOTE_NONNUMERIC,
}
LINE_TERM = {"LF": "\n", "CR": "\r", "CRLF": "\r\n", "os_default": linesep}
TIMESTAMP = dt.now().strftime("%Y-%m-%d--%H-%M-%S")
